# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tasks involving assembly and/or disassembly of bricks."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from absl import logging
from dm_control import composer
from dm_control.composer import initializers
from dm_control.composer import variation
from dm_control.composer.observation import observable
from dm_control.composer.variation import distributions
from dm_control.entities import props
from dm_control.manipulation.shared import arenas
from dm_control.manipulation.shared import cameras
from dm_control.manipulation.shared import constants
from dm_control.manipulation.shared import observations
from dm_control.manipulation.shared import registry
from dm_control.manipulation.shared import robots
from dm_control.manipulation.shared import tags
from dm_control.manipulation.shared import workspaces
from dm_control.mujoco.wrapper import mjbindings
from dm_control.utils import rewards
import numpy as np
from six.moves import range
from six.moves import zip

mjlib = mjbindings.mjlib


_BrickWorkspace = collections.namedtuple(
    '_BrickWorkspace',
    ['prop_bbox', 'tcp_bbox', 'goal_hint_pos', 'goal_hint_quat', 'arm_offset'])

# Ensures that the prop does not collide with the table during initialization.
_PROP_Z_OFFSET = 1e-6

_WORKSPACE = _BrickWorkspace(
    prop_bbox=workspaces.BoundingBox(
        lower=(-0.1, -0.1, _PROP_Z_OFFSET),
        upper=(0.1, 0.1, _PROP_Z_OFFSET)),
    tcp_bbox=workspaces.BoundingBox(
        lower=(-0.1, -0.1, 0.15),
        upper=(0.1, 0.1, 0.4)),
    goal_hint_pos=(0.2, 0.1, 0.),
    goal_hint_quat=(-0.38268343, 0., 0., 0.92387953),
    arm_offset=robots.ARM_OFFSET)

# Alpha value of the visual goal hint representing the goal state for each task.
_HINT_ALPHA = 0.75

# Distance thresholds for the shaping rewards for getting the top brick close
# to the bottom brick, and for 'clicking' them together.
_CLOSE_THRESHOLD = 0.01
_CLICK_THRESHOLD = 0.001

# Sequence of colors for the brick(s).
_COLOR_VALUES, _COLOR_NAMES = list(
    zip(
        ((1., 0., 0.), 'red'),
        ((0., 1., 0.), 'green'),
        ((0., 0., 1.), 'blue'),
        ((0., 1., 1.), 'cyan'),
        ((1., 0., 1.), 'magenta'),
        ((1., 1., 0.), 'yellow'),
    ))


class _Common(composer.Task):
  """Common components of brick tasks."""

  def __init__(self,
               arena,
               arm,
               hand,
               num_bricks,
               obs_settings,
               workspace,
               control_timestep):
    if not 2 <= num_bricks <= 6:
      raise ValueError('`num_bricks` must be between 2 and 6, got {}.'
                       .format(num_bricks))

    if num_bricks > 3:
      # The default values computed by MuJoCo's compiler are too small if there
      # are more than three stacked bricks, since each stacked pair generates
      # a large number of contacts. The values below are sufficient for up to
      # 6 stacked bricks.
      # TODO(b/78331644): It may be useful to log the size of `physics.model`
      #                   and `physics.data` after compilation to gauge the
      #                   impact of these changes on MuJoCo's memory footprint.
      arena.mjcf_model.size.nconmax = 400
      arena.mjcf_model.size.njmax = 1200

    self._arena = arena
    self._arm = arm
    self._hand = hand
    self._arm.attach(self._hand)
    self._arena.attach_offset(self._arm, offset=workspace.arm_offset)
    self.control_timestep = control_timestep

    # Add custom camera observable.
    self._task_observables = cameras.add_camera_observables(
        arena, obs_settings, cameras.FRONT_CLOSE)

    color_sequence = iter(_COLOR_VALUES)
    brick_obs_options = observations.make_options(
        obs_settings, observations.FREEPROP_OBSERVABLES)

    bricks = []
    brick_frames = []
    goal_hint_bricks = []
    for _ in range(num_bricks):
      color = next(color_sequence)
      brick = props.Duplo(color=color,
                          observable_options=brick_obs_options)
      brick_frames.append(arena.add_free_entity(brick))
      bricks.append(brick)

      # Translucent, contactless brick with no observables. These are used to
      # provide a visual hint representing the goal state for each task.
      hint_brick = props.Duplo(color=color)
      _hintify(hint_brick, alpha=_HINT_ALPHA)
      arena.attach(hint_brick)
      goal_hint_bricks.append(hint_brick)

    self._bricks = bricks
    self._brick_frames = brick_frames
    self._goal_hint_bricks = goal_hint_bricks

    # Position and quaternion for the goal hint.
    self._goal_hint_pos = workspace.goal_hint_pos
    self._goal_hint_quat = workspace.goal_hint_quat

    self._tcp_initializer = initializers.ToolCenterPointInitializer(
        self._hand, self._arm,
        position=distributions.Uniform(*workspace.tcp_bbox),
        quaternion=workspaces.DOWN_QUATERNION)

    # Add sites for visual debugging.
    workspaces.add_bbox_site(
        body=self.root_entity.mjcf_model.worldbody,
        lower=workspace.tcp_bbox.lower,
        upper=workspace.tcp_bbox.upper,
        rgba=constants.GREEN, name='tcp_spawn_area')

    workspaces.add_bbox_site(
        body=self.root_entity.mjcf_model.worldbody,
        lower=workspace.prop_bbox.lower,
        upper=workspace.prop_bbox.upper,
        rgba=constants.BLUE, name='prop_spawn_area')

  @property
  def task_observables(self):
    return self._task_observables

  @property
  def root_entity(self):
    return self._arena

  @property
  def arm(self):
    return self._arm

  @property
  def hand(self):
    return self._hand


class Stack(_Common):
  """Build a stack of Duplo bricks."""

  def __init__(self,
               arena,
               arm,
               hand,
               num_bricks,
               target_height,
               moveable_base,
               randomize_order,
               obs_settings,
               workspace,
               control_timestep):
    """Initializes a new `Stack` task.

    Args:
      arena: `composer.Entity` instance.
      arm: `robot_base.RobotArm` instance.
      hand: `robot_base.RobotHand` instance.
      num_bricks: The total number of bricks; must be between 2 and 6.
      target_height: The target number of bricks in the stack in order to get
        maximum reward. Must be between 2 and `num_bricks`.
      moveable_base: Boolean specifying whether or not the bottom brick should
        be moveable.
      randomize_order: Boolean specifying whether to randomize the desired order
        of bricks in the stack at the start of each episode.
      obs_settings: `observations.ObservationSettings` instance.
      workspace: A `_BrickWorkspace` instance.
      control_timestep: Float specifying the control timestep in seconds.

    Raises:
      ValueError: If `num_bricks` is not between 2 and 6, or if
        `target_height` is not between 2 and `num_bricks - 1`.
    """
    if not 2 <= target_height <= num_bricks:
      raise ValueError('`target_height` must be between 2 and {}, got {}.'
                       .format(num_bricks, target_height))

    super(Stack, self).__init__(arena=arena,
                                arm=arm,
                                hand=hand,
                                num_bricks=num_bricks,
                                obs_settings=obs_settings,
                                workspace=workspace,
                                control_timestep=control_timestep)

    self._moveable_base = moveable_base
    self._randomize_order = randomize_order
    self._target_height = target_height
    self._prop_bbox = workspace.prop_bbox

    # Shuffled at the start of each episode if `randomize_order` is True.
    self._desired_order = np.arange(target_height)

    # In the random order case, create a `prop_pose` observable that informs the
    # agent of the desired order.
    if randomize_order:
      desired_order_observable = observable.Generic(self._get_desired_order)
      desired_order_observable.configure(**obs_settings.prop_pose._asdict())
      self._task_observables['desired_order'] = desired_order_observable

  def _get_desired_order(self, physics):
    del physics  # Unused
    return self._desired_order.astype(np.double)

  def initialize_episode_mjcf(self, random_state):
    if self._randomize_order:
      self._desired_order = random_state.choice(
          len(self._bricks), size=self._target_height, replace=False)
      logging.info('Desired stack order (from bottom to top): [%s]',
                   ' '.join(_COLOR_NAMES[i] for i in self._desired_order))

    # If the base of the stack should be fixed, remove the freejoint for the
    # first brick (and ensure that all the others have freejoints).
    fixed_indices = [] if self._moveable_base else [self._desired_order[0]]
    _add_or_remove_freejoints(attachment_frames=self._brick_frames,
                              fixed_indices=fixed_indices)

    # We need to define the prop initializer for the bricks here rather than in
    # the `__init__`, since `PropPlacer` looks for freejoints on instantiation.
    self._brick_placer = initializers.PropPlacer(
        props=self._bricks,
        position=distributions.Uniform(*self._prop_bbox),
        quaternion=workspaces.uniform_z_rotation,
        settle_physics=True)

  def initialize_episode(self, physics, random_state):
    self._brick_placer(physics, random_state)
    self._hand.set_grasp(physics, close_factors=random_state.uniform())
    self._tcp_initializer(physics, random_state)
    # Arrange the goal hint bricks in the desired stack order.
    _build_stack(physics,
                 bricks=self._goal_hint_bricks,
                 base_pos=self._goal_hint_pos,
                 base_quat=self._goal_hint_quat,
                 order=self._desired_order,
                 random_state=random_state)

  def get_reward(self, physics):
    pairs = list(zip(self._desired_order[:-1], self._desired_order[1:]))
    pairwise_rewards = _get_pairwise_stacking_rewards(
        physics=physics, bricks=self._bricks, pairs=pairs)
    # The final reward is an average over the pairwise rewards.
    return np.mean(pairwise_rewards)


class Reassemble(_Common):
  """Disassemble a stack of Duplo bricks and reassemble it in another order."""

  def __init__(self,
               arena,
               arm,
               hand,
               num_bricks,
               randomize_initial_order,
               randomize_desired_order,
               obs_settings,
               workspace,
               control_timestep):
    """Initializes a new `Reassemble` task.

    Args:
      arena: `composer.Entity` instance.
      arm: `robot_base.RobotArm` instance.
      hand: `robot_base.RobotHand` instance.
      num_bricks: The total number of bricks; must be between 2 and 6.
      randomize_initial_order: Boolean specifying whether to randomize the
        initial order  of bricks in the stack at the start of each episode.
      randomize_desired_order: Boolean specifying whether to independently
        randomize the desired order of bricks in the stack at the start of each
        episode. By default the desired order will be the reverse of the initial
        order, with the exception of the base brick which is always the same as
        in the initial order since it is welded in place.
      obs_settings: `observations.ObservationSettings` instance.
      workspace: A `_BrickWorkspace` instance.
      control_timestep: Float specifying the control timestep in seconds.

    Raises:
      ValueError: If `num_bricks` is not between 2 and 6.
    """
    super(Reassemble, self).__init__(arena=arena,
                                     arm=arm,
                                     hand=hand,
                                     num_bricks=num_bricks,
                                     obs_settings=obs_settings,
                                     workspace=workspace,
                                     control_timestep=control_timestep)
    self._randomize_initial_order = randomize_initial_order
    self._randomize_desired_order = randomize_desired_order

    # Randomized at the start of each episode if `randomize_initial_order` is
    # True.
    self._initial_order = np.arange(num_bricks)

    # Randomized at the start of each episode if `randomize_desired_order` is
    # True.
    self._desired_order = self._initial_order.copy()
    self._desired_order[1:] = self._desired_order[-1:0:-1]

    # In the random order case, create a `prop_pose` observable that informs the
    # agent of the desired order.
    if randomize_desired_order:
      desired_order_observable = observable.Generic(self._get_desired_order)
      desired_order_observable.configure(**obs_settings.prop_pose._asdict())
      self._task_observables['desired_order'] = desired_order_observable

    # Distributions of positions and orientations for the base of the stack.
    self._base_pos = distributions.Uniform(*workspace.prop_bbox)
    self._base_quat = workspaces.uniform_z_rotation

  def _get_desired_order(self, physics):
    del physics  # Unused
    return self._desired_order.astype(np.double)

  def initialize_episode_mjcf(self, random_state):
    if self._randomize_initial_order:
      random_state.shuffle(self._initial_order)

    # The bottom brick will be fixed to the table, so it must be the same in
    # both the initial and desired order.
    self._desired_order[0] = self._initial_order[0]
    # By default the desired order of the other bricks is the opposite of their
    # initial order.
    self._desired_order[1:] = self._initial_order[-1:0:-1]

    if self._randomize_desired_order:
      random_state.shuffle(self._desired_order[1:])

    logging.info('Desired stack order (from bottom to top): [%s]',
                 ' '.join(_COLOR_NAMES[i] for i in self._desired_order))

    # Remove the freejoint from the bottom brick in the stack.
    _add_or_remove_freejoints(attachment_frames=self._brick_frames,
                              fixed_indices=[self._initial_order[0]])

  def initialize_episode(self, physics, random_state):
    # Build the initial stack.
    _build_stack(physics,
                 bricks=self._bricks,
                 base_pos=self._base_pos,
                 base_quat=self._base_quat,
                 order=self._initial_order,
                 random_state=random_state)
    # Arrange the goal hint bricks into a stack with the desired order.
    _build_stack(physics,
                 bricks=self._goal_hint_bricks,
                 base_pos=self._goal_hint_pos,
                 base_quat=self._goal_hint_quat,
                 order=self._desired_order,
                 random_state=random_state)
    self._hand.set_grasp(physics, close_factors=random_state.uniform())
    self._tcp_initializer(physics, random_state)

  def get_reward(self, physics):
    pairs = list(zip(self._desired_order[:-1], self._desired_order[1:]))
    # We set `close_coef=0.` because the coarse shaping reward causes problems
    # for this task (it means there is a strong disincentive to break up the
    # initial stack).
    pairwise_rewards = _get_pairwise_stacking_rewards(
        physics=physics,
        bricks=self._bricks,
        pairs=pairs,
        close_coef=0.)
    # The final reward is an average over the pairwise rewards.
    return np.mean(pairwise_rewards)


def _distance(pos1, pos2):
  diff = pos1 - pos2
  return sum(np.sqrt((diff * diff).sum(1)))


def _min_stud_to_hole_distance(physics, bottom_brick, top_brick):
  # Positions of the top left and bottom right studs on the `bottom_brick` and
  # the top left and bottom right holes on the `top_brick`.
  stud_pos = physics.bind(bottom_brick.studs[[0, -1], [0, -1]]).xpos
  hole_pos = physics.bind(top_brick.holes[[0, -1], [0, -1]]).xpos
  # Bricks are rotationally symmetric, so we compute top left -> top left and
  # top left -> bottom right distances and return whichever of these is smaller.
  dist1 = _distance(stud_pos, hole_pos)
  dist2 = _distance(stud_pos[::-1], hole_pos)
  return min(dist1, dist2)


def _get_pairwise_stacking_rewards(physics, bricks, pairs, close_coef=0.1):
  """Returns a vector of shaping reward components based on pairwise distances.

  Args:
    physics: An `mjcf.Physics` instance.
    bricks: A list of `composer.Entity` instances corresponding to bricks.
    pairs: A list of `(bottom_idx, top_idx)` tuples specifying which pairs of
      bricks should be measured.
    close_coef: Float specfying the relative weight given to the coarse-
      tolerance shaping component for getting the bricks close to one another
      (as opposed to the fine-tolerance component for clicking them together).

  Returns:
    A numpy array of size `len(pairs)` containing values in (0, 1], where
    1 corresponds to a stacked pair of bricks.
  """
  distances = []
  for bottom_idx, top_idx in pairs:
    bottom_brick = bricks[bottom_idx]
    top_brick = bricks[top_idx]
    distances.append(
        _min_stud_to_hole_distance(physics, bottom_brick, top_brick))
  distances = np.hstack(distances)

  # Coarse-tolerance component for bringing the holes close to the studs.
  close = rewards.tolerance(
      distances, bounds=(0, _CLOSE_THRESHOLD), margin=(_CLOSE_THRESHOLD * 10))

  # Fine-tolerance component for clicking the bricks together.
  clicked = rewards.tolerance(
      distances, bounds=(0, _CLICK_THRESHOLD), margin=_CLICK_THRESHOLD)

  # Weighted average of coarse and fine components for each pair of bricks.
  return np.average([close, clicked], weights=[close_coef, 1.], axis=0)


def _build_stack(physics, bricks, base_pos, base_quat, order, random_state):
  """Builds a stack of bricks.

  Args:
    physics: Instance of `mjcf.Physics`.
    bricks: Sequence of `composer.Entity` instances corresponding to bricks.
    base_pos: Position of the base brick in the stack; either a (3,) numpy array
      or a `variation.Variation` that yields such arrays.
    base_quat: Quaternion of the base brick in the stack; either a (4,) numpy
      array or a `variation.Variation` that yields such arrays.
    order: Sequence of indices specifying the order in which to stack the
      bricks.
    random_state: An `np.random.RandomState` instance.
  """
  base_pos = variation.evaluate(base_pos, random_state=random_state)
  base_quat = variation.evaluate(base_quat, random_state=random_state)
  bricks[order[0]].set_pose(physics, position=base_pos, quaternion=base_quat)
  for bottom_idx, top_idx in zip(order[:-1], order[1:]):
    bottom = bricks[bottom_idx]
    top = bricks[top_idx]
    stud_pos = physics.bind(bottom.studs[0, 0]).xpos
    _, quat = bottom.get_pose(physics)
    # The reward function treats top left -> top left and top left -> bottom
    # right configurations as identical, so the orientations of the bricks are
    # randomized so that 50% of the time the top brick is rotated 180 degrees
    # relative to the brick below.
    if random_state.rand() < 0.5:
      quat = quat.copy()
      axis = np.array([0., 0., 1.])
      angle = np.pi
      mjlib.mju_quatIntegrate(quat, axis, angle)
      hole_idx = (-1, -1)
    else:
      hole_idx = (0, 0)
    top.set_pose(physics, quaternion=quat)

    # Set the position of the top brick so that its holes line up with the studs
    # of the brick below.
    offset = physics.bind(top.holes[hole_idx]).xpos
    top_pos = stud_pos - offset
    top.set_pose(physics, position=top_pos)


def _add_or_remove_freejoints(attachment_frames, fixed_indices):
  """Adds or removes freejoints from props.

  Args:
    attachment_frames: A list of `mjcf.Elements` corresponding to attachment
      frames.
    fixed_indices: A list of indices of attachment frames that should be fixed
      to the world (i.e. have their freejoints removed). Freejoints will be
      added to all other elements in `attachment_frames` if they do not already
      possess them.
  """
  for i, frame in enumerate(attachment_frames):
    if i in fixed_indices:
      if frame.freejoint:
        frame.freejoint.remove()
    elif not frame.freejoint:
      frame.add('freejoint')


def _replace_alpha(rgba, alpha=0.3):
  new_rgba = rgba.copy()
  new_rgba[3] = alpha
  return new_rgba


def _hintify(entity, alpha=None):
  """Modifies an entity for use as a 'visual hint'.

  Contacts will be disabled for all geoms within the entity, and its bodies will
  be converted to "mocap" bodies (which are viewed as fixed from the perspective
  of the dynamics). The geom alpha values may also be overridden to render the
  geoms as translucent.

  Args:
    entity: A `composer.Entity`, modified in place.
    alpha: Optional float between 0 and 1, used to override the alpha values for
      all of the geoms in this entity.
  """
  for subentity in entity.iter_entities():
    # TODO(b/112084359): This assumes that all geoms either define explicit RGBA
    #                    values, or inherit from the top-level default. It will
    #                    not correctly handle more complicated hierarchies of
    #                    default classes.
    if (alpha is not None
        and subentity.mjcf_model.default.geom is not None
        and subentity.mjcf_model.default.geom.rgba is not None):
      subentity.mjcf_model.default.geom.rgba = _replace_alpha(
          subentity.mjcf_model.default.geom.rgba, alpha=alpha)
    for body in subentity.mjcf_model.find_all('body'):
      body.mocap = 'true'
    for geom in subentity.mjcf_model.find_all('geom'):
      if alpha is not None and geom.rgba is not None:
        geom.rgba = _replace_alpha(geom.rgba, alpha=alpha)
      geom.contype = 0
      geom.conaffinity = 0


def _stack(obs_settings, num_bricks, moveable_base, randomize_order,
           target_height=None):
  """Configure and instantiate a Stack task.

  Args:
    obs_settings: `observations.ObservationSettings` instance.
    num_bricks: The total number of bricks; must be between 2 and 6.
    moveable_base: Boolean specifying whether or not the bottom brick should
        be moveable.
    randomize_order: Boolean specifying whether to randomize the desired order
      of bricks in the stack at the start of each episode.
    target_height: The target number of bricks in the stack in order to get
      maximum reward. Must be between 2 and `num_bricks`. Defaults to
      `num_bricks`.

  Returns:
    An instance of `Stack`.
  """
  if target_height is None:
    target_height = num_bricks
  arena = arenas.Standard()
  arm = robots.make_arm(obs_settings=obs_settings)
  hand = robots.make_hand(obs_settings=obs_settings)
  return Stack(arena=arena,
               arm=arm,
               hand=hand,
               num_bricks=num_bricks,
               target_height=target_height,
               moveable_base=moveable_base,
               randomize_order=randomize_order,
               obs_settings=obs_settings,
               workspace=_WORKSPACE,
               control_timestep=constants.CONTROL_TIMESTEP)


@registry.add(tags.FEATURES)
def stack_2_bricks_features():
  return _stack(obs_settings=observations.PERFECT_FEATURES, num_bricks=2,
                moveable_base=False, randomize_order=False)


@registry.add(tags.VISION)
def stack_2_bricks_vision():
  return _stack(obs_settings=observations.VISION, num_bricks=2,
                moveable_base=False, randomize_order=False)


@registry.add(tags.FEATURES)
def stack_2_bricks_moveable_base_features():
  return _stack(obs_settings=observations.PERFECT_FEATURES, num_bricks=2,
                moveable_base=True, randomize_order=False)


@registry.add(tags.VISION)
def stack_2_bricks_moveable_base_vision():
  return _stack(obs_settings=observations.VISION, num_bricks=2,
                moveable_base=True, randomize_order=False)


@registry.add(tags.FEATURES)
def stack_3_bricks_features():
  return _stack(obs_settings=observations.PERFECT_FEATURES, num_bricks=3,
                moveable_base=False, randomize_order=False)


@registry.add(tags.VISION)
def stack_3_bricks_vision():
  return _stack(obs_settings=observations.VISION, num_bricks=3,
                moveable_base=False, randomize_order=False)


@registry.add(tags.FEATURES)
def stack_3_bricks_random_order_features():
  return _stack(obs_settings=observations.PERFECT_FEATURES, num_bricks=3,
                moveable_base=False, randomize_order=True)


@registry.add(tags.FEATURES)
def stack_2_of_3_bricks_random_order_features():
  return _stack(obs_settings=observations.PERFECT_FEATURES, num_bricks=3,
                moveable_base=False, randomize_order=True, target_height=2)


@registry.add(tags.VISION)
def stack_2_of_3_bricks_random_order_vision():
  return _stack(obs_settings=observations.VISION, num_bricks=3,
                moveable_base=False, randomize_order=True, target_height=2)


def _reassemble(obs_settings, num_bricks, randomize_initial_order,
                randomize_desired_order):
  """Configure and instantiate a `Reassemble` task.

  Args:
    obs_settings: `observations.ObservationSettings` instance.
    num_bricks: The total number of bricks; must be between 2 and 6.
    randomize_initial_order: Boolean specifying whether to randomize the
      initial order  of bricks in the stack at the start of each episode.
    randomize_desired_order: Boolean specifying whether to independently
      randomize the desired order of bricks in the stack at the start of each
      episode. By default the desired order will be the reverse of the initial
      order, with the exception of the base brick which is always the same as
      in the initial order since it is welded in place.

  Returns:
    An instance of `Reassemble`.
  """
  arena = arenas.Standard()
  arm = robots.make_arm(obs_settings=obs_settings)
  hand = robots.make_hand(obs_settings=obs_settings)
  return Reassemble(arena=arena,
                    arm=arm,
                    hand=hand,
                    num_bricks=num_bricks,
                    randomize_initial_order=randomize_initial_order,
                    randomize_desired_order=randomize_desired_order,
                    obs_settings=obs_settings,
                    workspace=_WORKSPACE,
                    control_timestep=constants.CONTROL_TIMESTEP)


@registry.add(tags.FEATURES)
def reassemble_3_bricks_fixed_order_features():
  return _reassemble(obs_settings=observations.PERFECT_FEATURES, num_bricks=3,
                     randomize_initial_order=False,
                     randomize_desired_order=False)


@registry.add(tags.VISION)
def reassemble_3_bricks_fixed_order_vision():
  return _reassemble(obs_settings=observations.VISION, num_bricks=3,
                     randomize_initial_order=False,
                     randomize_desired_order=False)


@registry.add(tags.FEATURES)
def reassemble_5_bricks_random_order_features():
  return _reassemble(obs_settings=observations.PERFECT_FEATURES, num_bricks=5,
                     randomize_initial_order=True,
                     randomize_desired_order=True)


@registry.add(tags.VISION)
def reassemble_5_bricks_random_order_vision():
  return _reassemble(obs_settings=observations.VISION, num_bricks=5,
                     randomize_initial_order=True,
                     randomize_desired_order=True)



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Quadruped Domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.mujoco.wrapper import mjbindings
from dm_control.rl import control
from dm_control.suite import base
from dm_control.suite import common
from dm_control.utils import containers
from dm_control.utils import rewards
from dm_control.utils import xml_tools

from lxml import etree
import numpy as np
from scipy import ndimage

enums = mjbindings.enums
mjlib = mjbindings.mjlib


_DEFAULT_TIME_LIMIT = 20
_CONTROL_TIMESTEP = .02

# Horizontal speeds above which the move reward is 1.
_RUN_SPEED = 5
_WALK_SPEED = 0.5

# Constants related to terrain generation.
_HEIGHTFIELD_ID = 0
_TERRAIN_SMOOTHNESS = 0.15  # 0.0: maximally bumpy; 1.0: completely smooth.
_TERRAIN_BUMP_SCALE = 2  # Spatial scale of terrain bumps (in meters).

# Named model elements.
_TOES = ['toe_front_left', 'toe_back_left', 'toe_back_right', 'toe_front_right']
_WALLS = ['wall_px', 'wall_py', 'wall_nx', 'wall_ny']

SUITE = containers.TaggedTasks()


def make_model(floor_size=None, terrain=False, rangefinders=False,
               walls_and_ball=False):
  """Returns the model XML string."""
  xml_string = common.read_model('quadruped.xml')
  parser = etree.XMLParser(remove_blank_text=True)
  mjcf = etree.XML(xml_string, parser)

  # Set floor size.
  if floor_size is not None:
    floor_geom = mjcf.find('.//geom[@name={!r}]'.format('floor'))
    floor_geom.attrib['size'] = '{} {} .5'.format(floor_size, floor_size)

  # Remove walls, ball and target.
  if not walls_and_ball:
    for wall in _WALLS:
      wall_geom = xml_tools.find_element(mjcf, 'geom', wall)
      wall_geom.getparent().remove(wall_geom)

    # Remove ball.
    ball_body = xml_tools.find_element(mjcf, 'body', 'ball')
    ball_body.getparent().remove(ball_body)

    # Remove target.
    target_site = xml_tools.find_element(mjcf, 'site', 'target')
    target_site.getparent().remove(target_site)

  # Remove terrain.
  if not terrain:
    terrain_geom = xml_tools.find_element(mjcf, 'geom', 'terrain')
    terrain_geom.getparent().remove(terrain_geom)

  # Remove rangefinders if they're not used, as range computations can be
  # expensive, especially in a scene with heightfields.
  if not rangefinders:
    rangefinder_sensors = mjcf.findall('.//rangefinder')
    for rf in rangefinder_sensors:
      rf.getparent().remove(rf)

  return etree.tostring(mjcf, pretty_print=True)


@SUITE.add()
def walk(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns the Walk task."""
  xml_string = make_model(floor_size=_DEFAULT_TIME_LIMIT * _WALK_SPEED)
  physics = Physics.from_xml_string(xml_string, common.ASSETS)
  task = Move(desired_speed=_WALK_SPEED, random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(physics, task, time_limit=time_limit,
                             control_timestep=_CONTROL_TIMESTEP,
                             **environment_kwargs)


@SUITE.add()
def run(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns the Run task."""
  xml_string = make_model(floor_size=_DEFAULT_TIME_LIMIT * _RUN_SPEED)
  physics = Physics.from_xml_string(xml_string, common.ASSETS)
  task = Move(desired_speed=_RUN_SPEED, random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(physics, task, time_limit=time_limit,
                             control_timestep=_CONTROL_TIMESTEP,
                             **environment_kwargs)


@SUITE.add()
def escape(time_limit=_DEFAULT_TIME_LIMIT, random=None,
           environment_kwargs=None):
  """Returns the Escape task."""
  xml_string = make_model(floor_size=40, terrain=True, rangefinders=True)
  physics = Physics.from_xml_string(xml_string, common.ASSETS)
  task = Escape(random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(physics, task, time_limit=time_limit,
                             control_timestep=_CONTROL_TIMESTEP,
                             **environment_kwargs)


@SUITE.add()
def fetch(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns the Fetch task."""
  xml_string = make_model(walls_and_ball=True)
  physics = Physics.from_xml_string(xml_string, common.ASSETS)
  task = Fetch(random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(physics, task, time_limit=time_limit,
                             control_timestep=_CONTROL_TIMESTEP,
                             **environment_kwargs)


class Physics(mujoco.Physics):
  """Physics simulation with additional features for the Quadruped domain."""

  def _reload_from_data(self, data):
    super(Physics, self)._reload_from_data(data)
    # Clear cached sensor names when the physics is reloaded.
    self._sensor_types_to_names = {}
    self._hinge_names = []

  def _get_sensor_names(self, *sensor_types):
    try:
      sensor_names = self._sensor_types_to_names[sensor_types]
    except KeyError:
      [sensor_ids] = np.where(np.in1d(self.model.sensor_type, sensor_types))
      sensor_names = [self.model.id2name(s_id, 'sensor') for s_id in sensor_ids]
      self._sensor_types_to_names[sensor_types] = sensor_names
    return sensor_names

  def torso_upright(self):
    """Returns the dot-product of the torso z-axis and the global z-axis."""
    return np.asarray(self.named.data.xmat['torso', 'zz'])

  def torso_velocity(self):
    """Returns the velocity of the torso, in the local frame."""
    return self.named.data.sensordata['velocimeter'].copy()

  def egocentric_state(self):
    """Returns the state without global orientation or position."""
    if not self._hinge_names:
      [hinge_ids] = np.nonzero(self.model.jnt_type ==
                               enums.mjtJoint.mjJNT_HINGE)
      self._hinge_names = [self.model.id2name(j_id, 'joint')
                           for j_id in hinge_ids]
    return np.hstack((self.named.data.qpos[self._hinge_names],
                      self.named.data.qvel[self._hinge_names],
                      self.data.act))

  def toe_positions(self):
    """Returns toe positions in egocentric frame."""
    torso_frame = self.named.data.xmat['torso'].reshape(3, 3)
    torso_pos = self.named.data.xpos['torso']
    torso_to_toe = self.named.data.xpos[_TOES] - torso_pos
    return torso_to_toe.dot(torso_frame)

  def force_torque(self):
    """Returns scaled force/torque sensor readings at the toes."""
    force_torque_sensors = self._get_sensor_names(enums.mjtSensor.mjSENS_FORCE,
                                                  enums.mjtSensor.mjSENS_TORQUE)
    return np.arcsinh(self.named.data.sensordata[force_torque_sensors])

  def imu(self):
    """Returns IMU-like sensor readings."""
    imu_sensors = self._get_sensor_names(enums.mjtSensor.mjSENS_GYRO,
                                         enums.mjtSensor.mjSENS_ACCELEROMETER)
    return self.named.data.sensordata[imu_sensors]

  def rangefinder(self):
    """Returns scaled rangefinder sensor readings."""
    rf_sensors = self._get_sensor_names(enums.mjtSensor.mjSENS_RANGEFINDER)
    rf_readings = self.named.data.sensordata[rf_sensors]
    no_intersection = -1.0
    return np.where(rf_readings == no_intersection, 1.0, np.tanh(rf_readings))

  def origin_distance(self):
    """Returns the distance from the origin to the workspace."""
    return np.asarray(np.linalg.norm(self.named.data.site_xpos['workspace']))

  def origin(self):
    """Returns origin position in the torso frame."""
    torso_frame = self.named.data.xmat['torso'].reshape(3, 3)
    torso_pos = self.named.data.xpos['torso']
    return -torso_pos.dot(torso_frame)

  def ball_state(self):
    """Returns ball position and velocity relative to the torso frame."""
    data = self.named.data
    torso_frame = data.xmat['torso'].reshape(3, 3)
    ball_rel_pos = data.xpos['ball'] - data.xpos['torso']
    ball_rel_vel = data.qvel['ball_root'][:3] - data.qvel['root'][:3]
    ball_rot_vel = data.qvel['ball_root'][3:]
    ball_state = np.vstack((ball_rel_pos, ball_rel_vel, ball_rot_vel))
    return ball_state.dot(torso_frame).ravel()

  def target_position(self):
    """Returns target position in torso frame."""
    torso_frame = self.named.data.xmat['torso'].reshape(3, 3)
    torso_pos = self.named.data.xpos['torso']
    torso_to_target = self.named.data.site_xpos['target'] - torso_pos
    return torso_to_target.dot(torso_frame)

  def ball_to_target_distance(self):
    """Returns horizontal distance from the ball to the target."""
    ball_to_target = (self.named.data.site_xpos['target'] -
                      self.named.data.xpos['ball'])
    return np.linalg.norm(ball_to_target[:2])

  def self_to_ball_distance(self):
    """Returns horizontal distance from the quadruped workspace to the ball."""
    self_to_ball = (self.named.data.site_xpos['workspace']
                    -self.named.data.xpos['ball'])
    return np.linalg.norm(self_to_ball[:2])


def _find_non_contacting_height(physics, orientation, x_pos=0.0, y_pos=0.0):
  """Find a height with no contacts given a body orientation.

  Args:
    physics: An instance of `Physics`.
    orientation: A quaternion.
    x_pos: A float. Position along global x-axis.
    y_pos: A float. Position along global y-axis.
  Raises:
    RuntimeError: If a non-contacting configuration has not been found after
    10,000 attempts.
  """
  z_pos = 0.0  # Start embedded in the floor.
  num_contacts = 1
  num_attempts = 0
  # Move up in 1cm increments until no contacts.
  while num_contacts > 0:
    try:
      with physics.reset_context():
        physics.named.data.qpos['root'][:3] = x_pos, y_pos, z_pos
        physics.named.data.qpos['root'][3:] = orientation
    except control.PhysicsError:
      # We may encounter a PhysicsError here due to filling the contact
      # buffer, in which case we simply increment the height and continue.
      pass
    num_contacts = physics.data.ncon
    z_pos += 0.01
    num_attempts += 1
    if num_attempts > 10000:
      raise RuntimeError('Failed to find a non-contacting configuration.')


def _common_observations(physics):
  """Returns the observations common to all tasks."""
  obs = collections.OrderedDict()
  obs['egocentric_state'] = physics.egocentric_state()
  obs['torso_velocity'] = physics.torso_velocity()
  obs['torso_upright'] = physics.torso_upright()
  obs['imu'] = physics.imu()
  obs['force_torque'] = physics.force_torque()
  return obs


def _upright_reward(physics, deviation_angle=0):
  """Returns a reward proportional to how upright the torso is.

  Args:
    physics: an instance of `Physics`.
    deviation_angle: A float, in degrees. The reward is 0 when the torso is
      exactly upside-down and 1 when the torso's z-axis is less than
      `deviation_angle` away from the global z-axis.
  """
  deviation = np.cos(np.deg2rad(deviation_angle))
  return rewards.tolerance(
      physics.torso_upright(),
      bounds=(deviation, float('inf')),
      sigmoid='linear',
      margin=1 + deviation,
      value_at_margin=0)


class Move(base.Task):
  """A quadruped task solved by moving forward at a designated speed."""

  def __init__(self, desired_speed, random=None):
    """Initializes an instance of `Move`.

    Args:
      desired_speed: A float. If this value is zero, reward is given simply
        for standing upright. Otherwise this specifies the horizontal velocity
        at which the velocity-dependent reward component is maximized.
      random: Optional, either a `numpy.random.RandomState` instance, an
        integer seed for creating a new `RandomState`, or None to select a seed
        automatically (default).
    """
    self._desired_speed = desired_speed
    super(Move, self).__init__(random=random)

  def initialize_episode(self, physics):
    """Sets the state of the environment at the start of each episode.

    Args:
      physics: An instance of `Physics`.

    """
    # Initial configuration.
    orientation = self.random.randn(4)
    orientation /= np.linalg.norm(orientation)
    _find_non_contacting_height(physics, orientation)
    super(Move, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns an observation to the agent."""
    return _common_observations(physics)

  def get_reward(self, physics):
    """Returns a reward to the agent."""

    # Move reward term.
    move_reward = rewards.tolerance(
        physics.torso_velocity()[0],
        bounds=(self._desired_speed, float('inf')),
        margin=self._desired_speed,
        value_at_margin=0.5,
        sigmoid='linear')

    return _upright_reward(physics) * move_reward


class Escape(base.Task):
  """A quadruped task solved by escaping a bowl-shaped terrain."""

  def initialize_episode(self, physics):
    """Sets the state of the environment at the start of each episode.

    Args:
      physics: An instance of `Physics`.

    """
    # Get heightfield resolution, assert that it is square.
    res = physics.model.hfield_nrow[_HEIGHTFIELD_ID]
    assert res == physics.model.hfield_ncol[_HEIGHTFIELD_ID]
    # Sinusoidal bowl shape.
    row_grid, col_grid = np.ogrid[-1:1:res*1j, -1:1:res*1j]
    radius = np.clip(np.sqrt(col_grid**2 + row_grid**2), .04, 1)
    bowl_shape = .5 - np.cos(2*np.pi*radius)/2
    # Random smooth bumps.
    terrain_size = 2 * physics.model.hfield_size[_HEIGHTFIELD_ID, 0]
    bump_res = int(terrain_size / _TERRAIN_BUMP_SCALE)
    bumps = self.random.uniform(_TERRAIN_SMOOTHNESS, 1, (bump_res, bump_res))
    smooth_bumps = ndimage.zoom(bumps, res / float(bump_res))
    # Terrain is elementwise product.
    terrain = bowl_shape * smooth_bumps
    start_idx = physics.model.hfield_adr[_HEIGHTFIELD_ID]
    physics.model.hfield_data[start_idx:start_idx+res**2] = terrain.ravel()
    super(Escape, self).initialize_episode(physics)

    # If we have a rendering context, we need to re-upload the modified
    # heightfield data.
    if physics.contexts:
      with physics.contexts.gl.make_current() as ctx:
        ctx.call(mjlib.mjr_uploadHField,
                 physics.model.ptr,
                 physics.contexts.mujoco.ptr,
                 _HEIGHTFIELD_ID)

    # Initial configuration.
    orientation = self.random.randn(4)
    orientation /= np.linalg.norm(orientation)
    _find_non_contacting_height(physics, orientation)

  def get_observation(self, physics):
    """Returns an observation to the agent."""
    obs = _common_observations(physics)
    obs['origin'] = physics.origin()
    obs['rangefinder'] = physics.rangefinder()
    return obs

  def get_reward(self, physics):
    """Returns a reward to the agent."""

    # Escape reward term.
    terrain_size = physics.model.hfield_size[_HEIGHTFIELD_ID, 0]
    escape_reward = rewards.tolerance(
        physics.origin_distance(),
        bounds=(terrain_size, float('inf')),
        margin=terrain_size,
        value_at_margin=0,
        sigmoid='linear')

    return _upright_reward(physics, deviation_angle=20) * escape_reward


class Fetch(base.Task):
  """A quadruped task solved by bringing a ball to the origin."""

  def initialize_episode(self, physics):
    """Sets the state of the environment at the start of each episode.

    Args:
      physics: An instance of `Physics`.

    """
    # Initial configuration, random azimuth and horizontal position.
    azimuth = self.random.uniform(0, 2*np.pi)
    orientation = np.array((np.cos(azimuth/2), 0, 0, np.sin(azimuth/2)))
    spawn_radius = 0.9 * physics.named.model.geom_size['floor', 0]
    x_pos, y_pos = self.random.uniform(-spawn_radius, spawn_radius, size=(2,))
    _find_non_contacting_height(physics, orientation, x_pos, y_pos)

    # Initial ball state.
    physics.named.data.qpos['ball_root'][:2] = self.random.uniform(
        -spawn_radius, spawn_radius, size=(2,))
    physics.named.data.qpos['ball_root'][2] = 2
    physics.named.data.qvel['ball_root'][:2] = 5*self.random.randn(2)
    super(Fetch, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns an observation to the agent."""
    obs = _common_observations(physics)
    obs['ball_state'] = physics.ball_state()
    obs['target_position'] = physics.target_position()
    return obs

  def get_reward(self, physics):
    """Returns a reward to the agent."""

    # Reward for moving close to the ball.
    arena_radius = physics.named.model.geom_size['floor', 0] * np.sqrt(2)
    workspace_radius = physics.named.model.site_size['workspace', 0]
    ball_radius = physics.named.model.geom_size['ball', 0]
    reach_reward = rewards.tolerance(
        physics.self_to_ball_distance(),
        bounds=(0, workspace_radius+ball_radius),
        sigmoid='linear',
        margin=arena_radius, value_at_margin=0)

    # Reward for bringing the ball to the target.
    target_radius = physics.named.model.site_size['target', 0]
    fetch_reward = rewards.tolerance(
        physics.ball_to_target_distance(),
        bounds=(0, target_radius),
        sigmoid='linear',
        margin=arena_radius, value_at_margin=0)

    reach_then_fetch = reach_reward * (0.5 + 0.5*fetch_reward)

    return _upright_reward(physics) * reach_then_fetch



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""An initializer that sets the pose of a hand's tool center point."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control import composer
from dm_control import mjcf
from dm_control.composer import variation
from dm_control.entities.manipulators import base
from six.moves import range


_REJECTION_SAMPLING_FAILED = (
    'Failed to find a valid initial configuration for the robot after '
    '{max_rejection_samples} TCP poses sampled and up to {max_ik_attempts} '
    'initial joint configurations per pose.')


class ToolCenterPointInitializer(composer.Initializer):
  """An initializer that sets the position of a hand's tool center point.

  This initializer calls the RobotArm's internal method to try and set the
  hand's TCP to a randomized Cartesian position within the specified bound.
  By default the initializer performs rejection sampling in order to avoid
  poses that result in "relevant collisions", which are defined as:

  * Collisions between links of the robot arm
  * Collisions between the arm and the hand
  * Collisions between either the arm or hand and an external body without a
    free joint
  """

  def __init__(self,
               hand,
               arm,
               position,
               quaternion=base.DOWN_QUATERNION,
               ignore_collisions=False,
               max_ik_attempts=10,
               max_rejection_samples=10):
    """Initializes this ToolCenterPointInitializer.

    Args:
      hand: Either a `base.RobotHand` instance or None, in which case
        `arm.wrist_site` is used as the TCP site in place of
        `hand.tool_center_point`.
      arm: A `base.RobotArm` instance.
      position: A single fixed Cartesian position, or a `Variation`
        object that generates Cartesian positions. If a fixed sequence of
        positions for multiple props is desired, use
        `variation.deterministic.Sequence`.
      quaternion: (optional) A single fixed unit quaternion, or a
        `composer.Variation` object that generates unit quaternions. If a fixed
        sequence of quaternions for ultiple props is desired, use
        `variation.deterministic.Sequence`.
      ignore_collisions: (optional) If True all collisions are ignored, i.e.
        rejection sampling is disabled.
      max_ik_attempts: (optional) Maximum number of attempts for the inverse
        kinematics solver to find a solution satisfying `target_pos` and
        `target_quat`. These are attempts per rejection sample. If more than
        one attempt is performed, the joint configuration will be randomized
        before the second trial. To avoid randomizing joint positions, set this
        parameter to 1.
      max_rejection_samples (optional): Maximum number of TCP target poses to
        sample while attempting to find a non-colliding configuration. For each
        sampled pose, up to `max_ik_attempts` may be performed in order to find
        an IK solution satisfying this pose.
    """
    super(ToolCenterPointInitializer, self).__init__()
    self._arm = arm
    self._hand = hand
    self._position = position
    self._quaternion = quaternion
    self._ignore_collisions = ignore_collisions
    self._max_ik_attempts = max_ik_attempts
    self._max_rejection_samples = max_rejection_samples

  def _has_relevant_collisions(self, physics):
    mjcf_root = self._arm.mjcf_model.root_model
    all_geoms = mjcf_root.find_all('geom')
    free_body_geoms = set()
    for body in mjcf_root.worldbody.get_children('body'):
      if mjcf.get_freejoint(body):
        free_body_geoms.update(body.find_all('geom'))

    arm_model = self._arm.mjcf_model
    hand_model = None
    if self._hand is not None:
      hand_model = self._hand.mjcf_model

    def is_robot(geom):
      return geom.root is arm_model or geom.root is hand_model

    def is_external_body_without_freejoint(geom):
      return not (is_robot(geom) or geom in free_body_geoms)

    for contact in physics.data.contact:
      geom_1 = all_geoms[contact.geom1]
      geom_2 = all_geoms[contact.geom2]
      if contact.dist > 0:
        # Ignore "contacts" with positive distance (i.e. not actually touching).
        continue
      if (
          # Include arm-arm and arm-hand self-collisions (but not hand-hand).
          (geom_1.root is arm_model and geom_2.root is arm_model) or
          (geom_1.root is arm_model and geom_2.root is hand_model) or
          (geom_1.root is hand_model and geom_2.root is arm_model) or
          # Include collisions between the arm or hand and an external body
          # provided that the external body does not have a freejoint.
          (is_robot(geom_1) and is_external_body_without_freejoint(geom_2)) or
          (is_external_body_without_freejoint(geom_1) and is_robot(geom_2))):
        return True
    return False

  def __call__(self, physics, random_state):
    """Sets initial tool center point pose via inverse kinematics.

    Args:
      physics: An `mjcf.Physics` instance.
      random_state: An `np.random.RandomState` instance.

    Raises:
      RuntimeError: If a collision-free pose could not be found within
        `max_ik_attempts`.
    """
    if self._hand is not None:
      target_site = self._hand.tool_center_point
    else:
      target_site = self._arm.wrist_site

    initial_qpos = physics.bind(self._arm.joints).qpos.copy()

    for _ in range(self._max_rejection_samples):
      target_pos = variation.evaluate(self._position,
                                      random_state=random_state)
      target_quat = variation.evaluate(self._quaternion,
                                       random_state=random_state)
      success = self._arm.set_site_to_xpos(
          physics=physics, random_state=random_state, site=target_site,
          target_pos=target_pos, target_quat=target_quat,
          max_ik_attempts=self._max_ik_attempts)

      if success:
        physics.forward()  # Recalculate contacts.
        if (self._ignore_collisions
            or not self._has_relevant_collisions(physics)):
          return

      # If IK failed to find a solution for this target pose, or if the solution
      # resulted in contacts, then reset the arm joints to their original
      # positions and try again with a new target.
      physics.bind(self._arm.joints).qpos = initial_qpos

    raise RuntimeError(_REJECTION_SAMPLING_FAILED.format(
        max_rejection_samples=self._max_rejection_samples,
        max_ik_attempts=self._max_ik_attempts))



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""RenderExecutor executes OpenGL rendering calls on an appropriate thread.

OpenGL calls must be made on the same thread as where an OpenGL context is
made current on. With GPU rendering, migrating OpenGL contexts between threads
can become expensive. We provide a thread-safe executor that maintains a
thread on which an OpenGL context can be kept permanently current, and any other
threads that wish to render with that context will have their rendering calls
offloaded to the dedicated thread.

For single-threaded applications, set the `DISABLE_RENDER_THREAD_OFFLOADING`
environment variable before launching the Python interpreter. This will
eliminate the overhead of unnecessary thread-switching.
"""

# pylint: disable=g-import-not-at-top
import os
_OFFLOAD = not bool(os.environ.get('DISABLE_RENDER_THREAD_OFFLOADING', ''))
del os

from dm_control._render.executor.render_executor import BaseRenderExecutor
from dm_control._render.executor.render_executor import OffloadingRenderExecutor
from dm_control._render.executor.render_executor import PassthroughRenderExecutor

_EXECUTORS = (PassthroughRenderExecutor, OffloadingRenderExecutor)

try:
  from dm_control._render.executor.native_mutex.render_executor import NativeMutexOffloadingRenderExecutor
  _EXECUTORS += (NativeMutexOffloadingRenderExecutor,)
except ImportError:
  NativeMutexOffloadingRenderExecutor = None

if _OFFLOAD:
  RenderExecutor = (  # pylint: disable=invalid-name
      NativeMutexOffloadingRenderExecutor or OffloadingRenderExecutor)
else:
  RenderExecutor = PassthroughRenderExecutor  # pylint: disable=invalid-name



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.utils.rewards."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.

from absl.testing import absltest
from absl.testing import parameterized

from dm_control.utils import rewards

import numpy as np


_INPUT_VECTOR_SIZE = 10
EPS = np.finfo(np.double).eps
INF = float("inf")


class ToleranceTest(parameterized.TestCase):

  @parameterized.parameters((0.5, 0.95), (1e12, 1-EPS), (1e12, EPS),
                            (EPS, 1-EPS), (EPS, EPS))
  def test_tolerance_sigmoid_parameterisation(self, margin, value_at_margin):
    actual = rewards.tolerance(x=margin, bounds=(0, 0), margin=margin,
                               value_at_margin=value_at_margin)
    self.assertAlmostEqual(actual, value_at_margin)

  @parameterized.parameters(("gaussian",), ("hyperbolic",), ("long_tail",),
                            ("cosine",), ("tanh_squared",), ("linear",),
                            ("quadratic",), ("reciprocal",))
  def test_tolerance_sigmoids(self, sigmoid):
    margins = [0.01, 1.0, 100, 10000]
    values_at_margin = [0.1, 0.5, 0.9]
    bounds_list = [(0, 0), (-1, 1), (-np.pi, np.pi), (-100, 100)]
    for bounds in bounds_list:
      for margin in margins:
        for value_at_margin in values_at_margin:
          upper_margin = bounds[1]+margin
          value = rewards.tolerance(x=upper_margin, bounds=bounds,
                                    margin=margin,
                                    value_at_margin=value_at_margin,
                                    sigmoid=sigmoid)
          self.assertAlmostEqual(value, value_at_margin, delta=np.sqrt(EPS))
          lower_margin = bounds[0]-margin
          value = rewards.tolerance(x=lower_margin, bounds=bounds,
                                    margin=margin,
                                    value_at_margin=value_at_margin,
                                    sigmoid=sigmoid)
          self.assertAlmostEqual(value, value_at_margin, delta=np.sqrt(EPS))

  @parameterized.parameters((-1, 0), (-0.5, 0.1), (0, 1), (0.5, 0.1), (1, 0))
  def test_tolerance_margin_loss_shape(self, x, expected):
    actual = rewards.tolerance(x=x, bounds=(0, 0), margin=0.5,
                               value_at_margin=0.1)
    self.assertAlmostEqual(actual, expected, delta=1e-3)

  def test_tolerance_vectorization(self):
    bounds = (-.1, .1)
    margin = 0.2
    x_array = np.random.randn(2, 3, 4)
    value_array = rewards.tolerance(x=x_array, bounds=bounds, margin=margin)
    self.assertEqual(x_array.shape, value_array.shape)
    for i, x in enumerate(x_array.ravel()):
      value = rewards.tolerance(x=x, bounds=bounds, margin=margin)
      self.assertEqual(value, value_array.ravel()[i])

  # pylint: disable=bad-whitespace
  @parameterized.parameters(
      # Exact target.
      (0,    (0, 0), 1),
      (EPS,  (0, 0), 0),
      (-EPS, (0, 0), 0),
      # Interval with one open end.
      (0,    (0, INF), 1),
      (EPS,  (0, INF), 1),
      (-EPS, (0, INF), 0),
      # Closed interval.
      (0,     (0, 1), 1),
      (EPS,   (0, 1), 1),
      (-EPS,  (0, 1), 0),
      (1,     (0, 1), 1),
      (1+EPS, (0, 1), 0))
  def test_tolerance_bounds(self, x, bounds, expected):
    actual = rewards.tolerance(x, bounds=bounds, margin=0)
    self.assertEqual(actual, expected)  # Should be exact, since margin == 0.

  def test_tolerance_incorrect_bounds_order(self):
    with self.assertRaisesWithLiteralMatch(
        ValueError, "Lower bound must be <= upper bound."):
      rewards.tolerance(0, bounds=(1, 0), margin=0.05)

  def test_tolerance_negative_margin(self):
    with self.assertRaisesWithLiteralMatch(
        ValueError, "`margin` must be non-negative."):
      rewards.tolerance(0, bounds=(0, 1), margin=-0.05)

  def test_tolerance_bad_value_at_margin(self):
    with self.assertRaisesWithLiteralMatch(
        ValueError, "`value_at_1` must be strictly between 0 and 1, got 0."):
      rewards.tolerance(0, bounds=(0, 1), margin=1, value_at_margin=0)

  def test_tolerance_unknown_sigmoid(self):
    with self.assertRaisesWithLiteralMatch(
        ValueError, "Unknown sigmoid type 'unsupported_sigmoid'."):
      rewards.tolerance(0, bounds=(0, 1), margin=.1,
                        sigmoid="unsupported_sigmoid")

if __name__ == "__main__":
  absltest.main()



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Viewer's windowing systems."""

from dm_control import _render

# pylint: disable=g-import-not-at-top
# pylint: disable=invalid-name

RenderWindow = None

try:
  from dm_control.viewer.gui import glfw_gui
  RenderWindow = glfw_gui.GlfwWindow
except ImportError:
  pass

if RenderWindow is None:

  def ErrorRenderWindow(*args, **kwargs):
    del args, kwargs
    raise ImportError(
        'Cannot create a window because no windowing system could be imported')
  RenderWindow = ErrorRenderWindow

del _render



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Module containing the Jaco robot class."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from dm_control import composer
from dm_control import mjcf
from dm_control.composer import define
from dm_control.composer.observation import observable
from dm_control.entities.manipulators import base
from dm_control.entities.manipulators.kinova import assets_path
import numpy as np

_JACO_ARM_XML_PATH = os.path.join(assets_path.KINOVA_ROOT, 'jaco_arm.xml')
_LARGE_JOINTS = ('joint_1', 'joint_2', 'joint_3')
_SMALL_JOINTS = ('joint_4', 'joint_5', 'joint_6')
_ALL_JOINTS = _LARGE_JOINTS + _SMALL_JOINTS
_WRIST_SITE = 'wristsite'

# These are the peak torque limits taken from Kinova's datasheet:
# https://www.kinovarobotics.com/sites/default/files/AS-ACT-KA58-KA75-SP-INT-EN%20201804-1.2%20%28KINOVA%E2%84%A2%20Actuator%20series%20KA75%2B%20KA-58%20Specifications%29.pdf
_LARGE_JOINT_MAX_TORQUE = 30.5
_SMALL_JOINT_MAX_TORQUE = 6.8

# On the real robot these limits are imposed by the actuator firmware. It's
# technically possible to exceed them via the low-level API, but this can reduce
# the lifetime of the actuators.
_LARGE_JOINT_MAX_VELOCITY = np.deg2rad(36.)
_SMALL_JOINT_MAX_VELOCITY = np.deg2rad(48.)

# The velocity actuator gain is a very rough estimate, and should be considered
# a placeholder for proper system identification.
_VELOCITY_GAIN = 500.


class JacoArm(base.RobotArm):
  """A composer entity representing a Jaco arm."""

  def _build(self, name=None):
    """Initializes the JacoArm.

    Args:
      name: String, the name of this robot. Used as a prefix in the MJCF name
        name attributes.
    """
    self._mjcf_root = mjcf.from_path(_JACO_ARM_XML_PATH)
    if name:
      self._mjcf_root.model = name
    # Find MJCF elements that will be exposed as attributes.
    self._joints = [self._mjcf_root.find('joint', name) for name in _ALL_JOINTS]
    self._wrist_site = self._mjcf_root.find('site', _WRIST_SITE)
    self._bodies = self.mjcf_model.find_all('body')
    # Add actuators.
    self._actuators = [_add_velocity_actuator(joint) for joint in self._joints]
    # Add torque sensors.
    self._joint_torque_sensors = [
        _add_torque_sensor(joint) for joint in self._joints]

  def _build_observables(self):
    return JacoArmObservables(self)

  @property
  def joints(self):
    """List of joint elements belonging to the arm."""
    return self._joints

  @property
  def actuators(self):
    """List of actuator elements belonging to the arm."""
    return self._actuators

  @property
  def joint_torque_sensors(self):
    """List of torque sensors for each joint belonging to the arm."""
    return self._joint_torque_sensors

  @property
  def wrist_site(self):
    """Wrist site of the arm (attachment point for the hand)."""
    return self._wrist_site

  @property
  def mjcf_model(self):
    """Returns the `mjcf.RootElement` object corresponding to this robot."""
    return self._mjcf_root


def _add_velocity_actuator(joint):
  """Adds a velocity actuator to a joint, returns the new MJCF element."""

  if joint.name in _LARGE_JOINTS:
    max_torque = _LARGE_JOINT_MAX_TORQUE
    max_velocity = _LARGE_JOINT_MAX_VELOCITY
  elif joint.name in _SMALL_JOINTS:
    max_torque = _SMALL_JOINT_MAX_TORQUE
    max_velocity = _SMALL_JOINT_MAX_VELOCITY
  else:
    raise ValueError('`joint.name` must be one of {}, got {!r}.'
                     .format(_ALL_JOINTS, joint.name))
  return joint.root.actuator.add(
      'velocity',
      joint=joint,
      name=joint.name,
      kv=_VELOCITY_GAIN,
      ctrllimited=True,
      ctrlrange=(-max_velocity, max_velocity),
      forcelimited=True,
      forcerange=(-max_torque, max_torque))


def _add_torque_sensor(joint):
  """Adds a torque sensor to a joint, returns the new MJCF element."""
  site = joint.parent.add(
      'site', size=[1e-3], group=composer.SENSOR_SITES_GROUP,
      name=joint.name+'_site')
  return joint.root.sensor.add('torque', site=site, name=joint.name+'_torque')


class JacoArmObservables(base.JointsObservables):
  """Jaco arm obserables."""

  @define.observable
  def joints_pos(self):
    # Because most of the Jaco arm joints are unlimited, we return the joint
    # angles as sine/cosine pairs so that the observations are bounded.
    def get_sin_cos_joint_angles(physics):
      joint_pos = physics.bind(self._entity.joints).qpos
      return np.vstack([np.sin(joint_pos), np.cos(joint_pos)]).T
    return observable.Generic(get_sin_cos_joint_angles)

  @define.observable
  def joints_torque(self):
    # MuJoCo's torque sensors are 3-axis, but we are only interested in torques
    # acting about the axis of rotation of the joint. We therefore project the
    # torques onto the joint axis.
    def get_torques(physics):
      torques = physics.bind(self._entity.joint_torque_sensors).sensordata
      joint_axes = physics.bind(self._entity.joints).axis
      return np.einsum('ij,ij->i', torques.reshape(-1, 3), joint_axes)
    return observable.Generic(get_torques)



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for `dm_control.mjcf.element`."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import hashlib
import itertools
import os
import sys
import traceback

# Internal dependencies.
from absl.testing import absltest
from absl.testing import parameterized
from dm_control import mjcf
from dm_control.mjcf import element
from dm_control.mjcf import namescope
from dm_control.mjcf import parser
from dm_control.mujoco.wrapper import util
import lxml
import numpy as np
import six
from six.moves import range

etree = lxml.etree

_ASSETS_DIR = os.path.join(os.path.dirname(__file__), 'test_assets')
_TEST_MODEL_XML = os.path.join(_ASSETS_DIR, 'test_model.xml')
_TEXTURE_PATH = os.path.join(_ASSETS_DIR, 'textures/deepmind.png')
_MESH_PATH = os.path.join(_ASSETS_DIR, 'meshes/cube.stl')
_MODEL_WITH_INCLUDE_PATH = os.path.join(_ASSETS_DIR, 'model_with_include.xml')

_MODEL_WITH_INVALID_FILENAMES = os.path.join(
    _ASSETS_DIR, 'model_with_invalid_filenames.xml')
_INCLUDED_WITH_INVALID_FILENAMES = os.path.join(
    _ASSETS_DIR, 'included_with_invalid_filenames.xml')
_MODEL_WITH_NAMELESS_ASSETS = os.path.join(
    _ASSETS_DIR, 'model_with_nameless_assets.xml')


class ElementTest(parameterized.TestCase):

  def assertIsSame(self, mjcf_model, other):
    self.assertTrue(mjcf_model.is_same_as(other))
    self.assertTrue(other.is_same_as(mjcf_model))

  def assertIsNotSame(self, mjcf_model, other):
    self.assertFalse(mjcf_model.is_same_as(other))
    self.assertFalse(other.is_same_as(mjcf_model))

  def assertHasAttr(self, obj, attrib):
    self.assertTrue(hasattr(obj, attrib))

  def assertNotHasAttr(self, obj, attrib):
    self.assertFalse(hasattr(obj, attrib))

  def _test_properties(self, mjcf_element, parent, root, recursive=False):
    self.assertEqual(mjcf_element.tag, mjcf_element.spec.name)
    self.assertEqual(mjcf_element.parent, parent)
    self.assertEqual(mjcf_element.root, root)
    self.assertEqual(mjcf_element.namescope, root.namescope)
    for child_name, child_spec in six.iteritems(mjcf_element.spec.children):
      if not (child_spec.repeated or child_spec.on_demand):
        child = getattr(mjcf_element, child_name)
        self.assertEqual(child.tag, child_name)
        self.assertEqual(child.spec, child_spec)
        if recursive:
          self._test_properties(child, parent=mjcf_element,
                                root=root, recursive=True)

  def testAttributeError(self):
    mjcf_model = element.RootElement(model='test')
    mjcf_model.worldbody._spec = None
    try:
      _ = mjcf_model.worldbody.tag
    except AttributeError:
      _, err, tb = sys.exc_info()
    else:
      self.fail('AttributeError was not raised.')
    # Test that the error comes from the fact that we've set `_spec = None`.
    self.assertEqual(str(err),
                     '\'NoneType\' object has no attribute \'name\'')
    _, _, func_name, _ = traceback.extract_tb(tb)[-1]
    # Test that the error comes from the `root` property, not `__getattr__`.
    self.assertEqual(func_name, 'tag')

  def testProperties(self):
    mujoco = element.RootElement(model='test')
    self.assertIsInstance(mujoco.namescope, namescope.NameScope)
    self._test_properties(mujoco, parent=None, root=mujoco, recursive=True)

  def _test_attributes(self, mjcf_element,
                       expected_values=None, recursive=False):
    attributes = mjcf_element.get_attributes()
    self.assertNotIn('class', attributes)
    for attribute_name in six.iterkeys(mjcf_element.spec.attributes):
      if attribute_name == 'class':
        attribute_name = 'dclass'
      self.assertHasAttr(mjcf_element, attribute_name)
      self.assertIn(attribute_name, dir(mjcf_element))
      attribute_value = getattr(mjcf_element, attribute_name)
      if attribute_value is not None:
        self.assertIn(attribute_name, attributes)
      else:
        self.assertNotIn(attribute_name, attributes)
      if expected_values:
        if attribute_name in expected_values:
          expected_value = expected_values[attribute_name]
          np.testing.assert_array_equal(attribute_value, expected_value)
        else:
          self.assertIsNone(attribute_value)
    if recursive:
      for child in mjcf_element.all_children():
        self._test_attributes(child, recursive=True)

  def testAttributes(self):
    mujoco = element.RootElement(model='test')
    mujoco.default.dclass = 'main'
    self._test_attributes(mujoco, recursive=True)

  def _test_children(self, mjcf_element, recursive=False):
    children = mjcf_element.all_children()
    for child_name, child_spec in six.iteritems(mjcf_element.spec.children):
      if not (child_spec.repeated or child_spec.on_demand):
        self.assertHasAttr(mjcf_element, child_name)
        self.assertIn(child_name, dir(mjcf_element))
        child = getattr(mjcf_element, child_name)
        self.assertIn(child, children)
        with six.assertRaisesRegex(
            self, AttributeError, 'can\'t set attribute'):
          setattr(mjcf_element, child_name, 'value')
        if recursive:
          self._test_children(child, recursive=True)

  def testChildren(self):
    mujoco = element.RootElement(model='test')
    self._test_children(mujoco, recursive=True)

  def testInvalidAttr(self):
    mujoco = element.RootElement(model='test')
    invalid_attrib_name = 'foobar'
    def test_invalid_attr_recursively(mjcf_element):
      self.assertNotHasAttr(mjcf_element, invalid_attrib_name)
      self.assertNotIn(invalid_attrib_name, dir(mjcf_element))
      with six.assertRaisesRegex(
          self, AttributeError, 'object has no attribute'):
        getattr(mjcf_element, invalid_attrib_name)
      with six.assertRaisesRegex(
          self, AttributeError, 'can\'t set attribute'):
        setattr(mjcf_element, invalid_attrib_name, 'value')
      with six.assertRaisesRegex(
          self, AttributeError, 'object has no attribute'):
        delattr(mjcf_element, invalid_attrib_name)
      for child in mjcf_element.all_children():
        test_invalid_attr_recursively(child)
    test_invalid_attr_recursively(mujoco)

  def testAdd(self):
    mujoco = element.RootElement(model='test')

    # repeated elements
    body_foo_attributes = dict(name='foo', pos=[0, 1, 0], quat=[0, 1, 0, 0])
    body_foo = mujoco.worldbody.add('body', **body_foo_attributes)
    self.assertEqual(body_foo.tag, 'body')
    joint_foo_attributes = dict(name='foo', type='free')
    joint_foo = body_foo.add('joint', **joint_foo_attributes)
    self.assertEqual(joint_foo.tag, 'joint')
    self._test_properties(body_foo, parent=mujoco.worldbody, root=mujoco)
    self._test_attributes(body_foo, expected_values=body_foo_attributes)
    self._test_children(body_foo)
    self._test_properties(joint_foo, parent=body_foo, root=mujoco)
    self._test_attributes(joint_foo, expected_values=joint_foo_attributes)
    self._test_children(joint_foo)

    # non-repeated, on-demand elements
    self.assertIsNone(body_foo.inertial)
    body_foo_inertial_attributes = dict(mass=1.0, pos=[0, 0, 0])
    body_foo_inertial = body_foo.add('inertial', **body_foo_inertial_attributes)
    self._test_properties(body_foo_inertial, parent=body_foo, root=mujoco)
    self._test_attributes(body_foo_inertial,
                          expected_values=body_foo_inertial_attributes)
    self._test_children(body_foo_inertial)

    with six.assertRaisesRegex(
        self, ValueError, '<inertial> child already exists'):
      body_foo.add('inertial', **body_foo_inertial_attributes)

    # non-repeated, non-on-demand elements
    with six.assertRaisesRegex(
        self, ValueError, '<compiler> child already exists'):
      mujoco.add('compiler')
    self.assertIsNotNone(mujoco.compiler)
    with six.assertRaisesRegex(
        self, ValueError, '<default> child already exists'):
      mujoco.add('default')
    self.assertIsNotNone(mujoco.default)

  def testAddWithInvalidAttribute(self):
    mujoco = element.RootElement(model='test')
    with six.assertRaisesRegex(
        self, AttributeError, 'not a valid attribute'):
      mujoco.worldbody.add('body', name='foo', invalid_attribute='some_value')
    self.assertFalse(mujoco.worldbody.body)
    self.assertIsNone(mujoco.worldbody.find('body', 'foo'))

  def testSameness(self):
    mujoco = element.RootElement(model='test')

    body_1 = mujoco.worldbody.add('body', pos=[0, 1, 2], quat=[0, 1, 0, 1])
    site_1 = body_1.add('site', pos=[0, 1, 2], quat=[0, 1, 0, 1])
    geom_1 = body_1.add('geom', pos=[0, 1, 2], quat=[0, 1, 0, 1])

    for elem in (body_1, site_1, geom_1):
      self.assertIsSame(elem, elem)

    # strict ordering NOT required: adding geom and site is different order
    body_2 = mujoco.worldbody.add('body', pos=[0, 1, 2], quat=[0, 1, 0, 1])
    geom_2 = body_2.add('geom', pos=[0, 1, 2], quat=[0, 1, 0, 1])
    site_2 = body_2.add('site', pos=[0, 1, 2], quat=[0, 1, 0, 1])

    elems_1 = (body_1, site_1, geom_1)
    elems_2 = (body_2, site_2, geom_2)
    for i, j in itertools.product(range(len(elems_1)), range(len(elems_2))):
      if i == j:
        self.assertIsSame(elems_1[i], elems_2[j])
      else:
        self.assertIsNotSame(elems_1[i], elems_2[j])

    # on-demand child
    body_1.add('inertial', pos=[0, 0, 0], mass=1)
    self.assertIsNotSame(body_1, body_2)

    body_2.add('inertial', pos=[0, 0, 0], mass=1)
    self.assertIsSame(body_1, body_2)

    # different number of children
    subbody_1 = body_1.add('body', pos=[0, 0, 1])
    self.assertIsNotSame(body_1, body_2)

    # attribute mismatch
    subbody_2 = body_2.add('body')
    self.assertIsNotSame(subbody_1, subbody_2)
    self.assertIsNotSame(body_1, body_2)

    subbody_2.pos = [0, 0, 1]
    self.assertIsSame(subbody_1, subbody_2)
    self.assertIsSame(body_1, body_2)

    # grandchild attribute mismatch
    subbody_1.add('joint', type='hinge')
    subbody_2.add('joint', type='ball')
    self.assertIsNotSame(body_1, body_2)

  def testTendonSameness(self):
    mujoco = element.RootElement(model='test')

    spatial_1 = mujoco.tendon.add('spatial')
    spatial_1.add('site', site='foo')
    spatial_1.add('geom', geom='bar')

    spatial_2 = mujoco.tendon.add('spatial')
    spatial_2.add('site', site='foo')
    spatial_2.add('geom', geom='bar')

    self.assertIsSame(spatial_1, spatial_2)

    # strict ordering is required
    spatial_3 = mujoco.tendon.add('spatial')
    spatial_3.add('site', site='foo')
    spatial_3.add('geom', geom='bar')

    spatial_4 = mujoco.tendon.add('spatial')
    spatial_4.add('geom', geom='bar')
    spatial_4.add('site', site='foo')

    self.assertIsNotSame(spatial_3, spatial_4)

  def testCopy(self):
    mujoco = parser.from_path(_TEST_MODEL_XML)
    self.assertIsSame(mujoco, mujoco)

    copy_mujoco = copy.copy(mujoco)
    copy_mujoco.model = 'copied_model'
    self.assertIsSame(copy_mujoco, mujoco)
    self.assertNotEqual(copy_mujoco, mujoco)

    deepcopy_mujoco = copy.deepcopy(mujoco)
    deepcopy_mujoco.model = 'deepcopied_model'
    self.assertIsSame(deepcopy_mujoco, mujoco)
    self.assertNotEqual(deepcopy_mujoco, mujoco)

    self.assertIsSame(deepcopy_mujoco, copy_mujoco)
    self.assertNotEqual(deepcopy_mujoco, copy_mujoco)

  def testWorldBodyFullIdentifier(self):
    mujoco = parser.from_path(_TEST_MODEL_XML)
    mujoco.model = 'model'
    self.assertEqual(mujoco.worldbody.full_identifier, 'world')

    submujoco = copy.copy(mujoco)
    submujoco.model = 'submodel'
    self.assertEqual(submujoco.worldbody.full_identifier, 'world')

    mujoco.attach(submujoco)
    self.assertEqual(mujoco.worldbody.full_identifier, 'world')
    self.assertEqual(submujoco.worldbody.full_identifier, 'submodel/')

    self.assertNotIn('name', mujoco.worldbody.to_xml_string(self_only=True))
    self.assertNotIn('name', submujoco.worldbody.to_xml_string(self_only=True))

  def testAttach(self):
    mujoco = parser.from_path(_TEST_MODEL_XML)
    mujoco.model = 'model'

    submujoco = copy.copy(mujoco)
    submujoco.model = 'submodel'

    subsubmujoco = copy.copy(mujoco)
    subsubmujoco.model = 'subsubmodel'

    with six.assertRaisesRegex(
        self, ValueError, 'Cannot merge a model to itself'):
      mujoco.attach(mujoco)

    attachment_site = submujoco.find('site', 'attachment')
    attachment_site.attach(subsubmujoco)
    subsubmodel_frame = submujoco.find('attachment_frame', 'subsubmodel')
    for attribute_name in ('pos', 'axisangle', 'xyaxes', 'zaxis', 'euler'):
      np.testing.assert_array_equal(
          getattr(subsubmodel_frame, attribute_name),
          getattr(attachment_site, attribute_name))
    self._test_properties(subsubmodel_frame,
                          parent=attachment_site.parent, root=submujoco)
    self.assertEqual(
        subsubmodel_frame.to_xml_string().split('\n')[0],
        '<body '
        'pos="0.10000000000000001 0.10000000000000001 0.10000000000000001" '
        'quat="0 1 0 0" '
        'name="subsubmodel/">')
    self.assertEqual(subsubmodel_frame.all_children(),
                     subsubmujoco.worldbody.all_children())

    with six.assertRaisesRegex(self, ValueError, 'already attached elsewhere'):
      mujoco.attach(subsubmujoco)

    with six.assertRaisesRegex(self, ValueError, 'Expected a mjcf.RootElement'):
      mujoco.attach(submujoco.contact)

    submujoco.option.flag.gravity = 'enable'
    with six.assertRaisesRegex(self, ValueError,
                               'Conflicting values for attribute `gravity`'):
      mujoco.attach(submujoco)
    submujoco.option.flag.gravity = 'disable'

    mujoco.attach(submujoco)
    self.assertEqual(subsubmujoco.parent_model, submujoco)
    self.assertEqual(submujoco.parent_model, mujoco)
    self.assertEqual(subsubmujoco.root_model, mujoco)
    self.assertEqual(submujoco.root_model, mujoco)

    self.assertEqual(submujoco.full_identifier, 'submodel/')
    self.assertEqual(subsubmujoco.full_identifier, 'submodel/subsubmodel/')

    merged_children = ('contact', 'actuator')
    for child_name in merged_children:
      for grandchild in getattr(submujoco, child_name).all_children():
        self.assertIn(grandchild, getattr(mujoco, child_name).all_children())
      for grandchild in getattr(subsubmujoco, child_name).all_children():
        self.assertIn(grandchild, getattr(mujoco, child_name).all_children())
        self.assertIn(grandchild, getattr(submujoco, child_name).all_children())

    base_contact_content = (
        '<exclude name="{0}exclude" body1="{0}b_0" body2="{0}b_1"/>')
    self.assertEqual(
        mujoco.contact.to_xml_string(pretty_print=False),
        '<contact>' +
        base_contact_content.format('') +
        base_contact_content.format('submodel/') +
        base_contact_content.format('submodel/subsubmodel/') +
        '</contact>')

    actuators_template = (
        '<velocity name="{1}b_0_0" class="{0}" joint="{1}b_0_0"/>'
        '<velocity name="{1}b_1_0" class="{0}" joint="{1}b_1_0"/>')
    self.assertEqual(
        mujoco.actuator.to_xml_string(pretty_print=False),
        '<actuator>' +
        actuators_template.format('/', '') +
        actuators_template.format('submodel/', 'submodel/') +
        actuators_template.format('submodel/subsubmodel/',
                                  'submodel/subsubmodel/') +
        '</actuator>')

    self.assertEqual(mujoco.default.full_identifier, '/')
    self.assertEqual(mujoco.default.default[0].full_identifier, 'big_and_green')
    self.assertEqual(submujoco.default.full_identifier, 'submodel/')
    self.assertEqual(submujoco.default.default[0].full_identifier,
                     'submodel/big_and_green')
    self.assertEqual(subsubmujoco.default.full_identifier,
                     'submodel/subsubmodel/')
    self.assertEqual(subsubmujoco.default.default[0].full_identifier,
                     'submodel/subsubmodel/big_and_green')
    default_xml_lines = (mujoco.default.to_xml_string(pretty_print=False)
                         .replace('><', '>><<').split('><'))
    self.assertEqual(default_xml_lines[0], '<default>')
    self.assertEqual(default_xml_lines[1], '<default class="/">')
    self.assertEqual(default_xml_lines[4], '<default class="big_and_green">')
    self.assertEqual(default_xml_lines[6], '</default>')
    self.assertEqual(default_xml_lines[7], '</default>')
    self.assertEqual(default_xml_lines[8], '<default class="submodel/">')
    self.assertEqual(default_xml_lines[11],
                     '<default class="submodel/big_and_green">')
    self.assertEqual(default_xml_lines[13], '</default>')
    self.assertEqual(default_xml_lines[14], '</default>')
    self.assertEqual(default_xml_lines[15],
                     '<default class="submodel/subsubmodel/">')
    self.assertEqual(default_xml_lines[18],
                     '<default class="submodel/subsubmodel/big_and_green">')
    self.assertEqual(default_xml_lines[-3], '</default>')
    self.assertEqual(default_xml_lines[-2], '</default>')
    self.assertEqual(default_xml_lines[-1], '</default>')

  def testDetach(self):
    root = parser.from_path(_TEST_MODEL_XML)
    root.model = 'model'

    submodel = copy.copy(root)
    submodel.model = 'submodel'

    unattached_xml_1 = root.to_xml_string()
    root.attach(submodel)
    attached_xml_1 = root.to_xml_string()

    submodel.detach()
    unattached_xml_2 = root.to_xml_string()
    root.attach(submodel)
    attached_xml_2 = root.to_xml_string()

    self.assertEqual(unattached_xml_1, unattached_xml_2)
    self.assertEqual(attached_xml_1, attached_xml_2)

  def testRenameAttachedModel(self):
    root = parser.from_path(_TEST_MODEL_XML)
    root.model = 'model'

    submodel = copy.copy(root)
    submodel.model = 'submodel'
    geom = submodel.worldbody.add(
        'geom', name='geom', type='sphere', size=[0.1])

    frame = root.attach(submodel)
    submodel.model = 'renamed'
    self.assertEqual(frame.full_identifier, 'renamed/')
    self.assertIsSame(root.find('geom', 'renamed/geom'), geom)

  def testAttachmentFrames(self):
    mujoco = parser.from_path(_TEST_MODEL_XML)
    mujoco.model = 'model'

    submujoco = copy.copy(mujoco)
    submujoco.model = 'submodel'

    subsubmujoco = copy.copy(mujoco)
    subsubmujoco.model = 'subsubmodel'

    attachment_site = submujoco.find('site', 'attachment')
    attachment_site.attach(subsubmujoco)
    mujoco.attach(submujoco)

    # attachments directly on worldbody can have a <freejoint>
    submujoco_frame = mujoco.find('attachment_frame', 'submodel')
    self.assertStartsWith(submujoco_frame.to_xml_string(pretty_print=False),
                          '<body name="submodel/">')
    self.assertEqual(submujoco_frame.full_identifier, 'submodel/')
    free_joint = submujoco_frame.add('freejoint')
    self.assertEqual(free_joint.to_xml_string(pretty_print=False),
                     '<freejoint name="submodel/"/>')
    self.assertEqual(free_joint.full_identifier, 'submodel/')

    # attachments elsewhere cannot have a <freejoint>
    subsubmujoco_frame = submujoco.find('attachment_frame', 'subsubmodel')
    subsubmujoco_frame_xml = subsubmujoco_frame.to_xml_string(
        pretty_print=False, prefix_root=mujoco.namescope)
    self.assertStartsWith(
        subsubmujoco_frame_xml,
        '<body '
        'pos="0.10000000000000001 0.10000000000000001 0.10000000000000001" '
        'quat="0 1 0 0" '
        'name="submodel/subsubmodel/">')
    self.assertEqual(subsubmujoco_frame.full_identifier,
                     'submodel/subsubmodel/')
    with six.assertRaisesRegex(self, AttributeError, 'not a valid child'):
      subsubmujoco_frame.add('freejoint')
    hinge_joint = subsubmujoco_frame.add('joint', type='hinge', axis=[1, 2, 3])
    hinge_joint_xml = hinge_joint.to_xml_string(
        pretty_print=False, prefix_root=mujoco.namescope)
    self.assertEqual(
        hinge_joint_xml,
        '<joint class="submodel/" type="hinge" axis="1 2 3" '
        'name="submodel/subsubmodel/"/>')
    self.assertEqual(hinge_joint.full_identifier, 'submodel/subsubmodel/')

  def testDuplicateAttachmentFrameJointIdentifiers(self):
    mujoco = parser.from_path(_TEST_MODEL_XML)
    mujoco.model = 'model'

    submujoco_1 = copy.copy(mujoco)
    submujoco_1.model = 'submodel_1'

    submujoco_2 = copy.copy(mujoco)
    submujoco_2.model = 'submodel_2'

    frame_1 = mujoco.attach(submujoco_1)
    frame_2 = mujoco.attach(submujoco_2)

    joint_1 = frame_1.add('joint', type='slide', name='root_x', axis=[1, 0, 0])
    joint_2 = frame_2.add('joint', type='slide', name='root_x', axis=[1, 0, 0])

    self.assertEqual(joint_1.full_identifier, 'submodel_1/root_x/')
    self.assertEqual(joint_2.full_identifier, 'submodel_2/root_x/')

  def testAttachmentFrameReference(self):
    root_1 = mjcf.RootElement('model_1')
    root_2 = mjcf.RootElement('model_2')
    root_2_frame = root_1.attach(root_2)
    sensor = root_1.sensor.add(
        'framelinacc', name='root_2', objtype='body', objname=root_2_frame)
    self.assertEqual(
        sensor.to_xml_string(pretty_print=False),
        '<framelinacc name="root_2" objtype="body" objname="model_2/"/>')

  def testAttachmentFrameChildReference(self):
    root_1 = mjcf.RootElement('model_1')
    root_2 = mjcf.RootElement('model_2')
    root_2_frame = root_1.attach(root_2)
    root_2_joint = root_2_frame.add(
        'joint', name='root_x', type='slide', axis=[1, 0, 0])
    actuator = root_1.actuator.add(
        'position', name='root_x', joint=root_2_joint)
    self.assertEqual(
        actuator.to_xml_string(pretty_print=False),
        '<position name="root_x" class="/" joint="model_2/root_x/"/>')

  def testDeletion(self):
    mujoco = parser.from_path(_TEST_MODEL_XML)
    mujoco.model = 'model'

    submujoco = copy.copy(mujoco)
    submujoco.model = 'submodel'

    subsubmujoco = copy.copy(mujoco)
    subsubmujoco.model = 'subsubmodel'

    submujoco.find('site', 'attachment').attach(subsubmujoco)
    mujoco.attach(submujoco)

    with six.assertRaisesRegex(self, ValueError,
                               r'use remove\(affect_attachments=True\)'):
      del mujoco.option

    mujoco.option.remove(affect_attachments=True)
    for root in (mujoco, submujoco, subsubmujoco):
      self.assertIsNotNone(root.option.flag)
      self.assertEqual(
          root.option.to_xml_string(pretty_print=False), '<option/>')
      self.assertIsNotNone(root.option.flag)
      self.assertEqual(
          root.option.flag.to_xml_string(pretty_print=False), '<flag/>')

    with six.assertRaisesRegex(self, ValueError,
                               r'use remove\(affect_attachments=True\)'):
      del mujoco.contact

    mujoco.contact.remove(affect_attachments=True)
    for root in (mujoco, submujoco, subsubmujoco):
      self.assertEqual(
          root.contact.to_xml_string(pretty_print=False), '<contact/>')

    b_0 = mujoco.find('body', 'b_0')
    b_0_inertial = b_0.inertial
    self.assertEqual(b_0_inertial.mass, 1)
    self.assertIsNotNone(b_0.inertial)
    del b_0.inertial
    self.assertIsNone(b_0.inertial)

  def testRemoveElementWithRequiredAttribute(self):
    root = mjcf.RootElement()
    body = root.worldbody.add('body')
    # `objtype` is a required attribute.
    sensor = root.sensor.add('framepos', objtype='body', objname=body)
    self.assertIn(sensor, root.sensor.all_children())
    sensor.remove()
    self.assertNotIn(sensor, root.sensor.all_children())

  def testRemoveWithChildren(self):
    root = mjcf.RootElement()
    body = root.worldbody.add('body')
    subbodies = []
    for _ in range(5):
      subbodies.append(body.add('body'))
    body.remove()
    for subbody in subbodies:
      self.assertTrue(subbody.is_removed)

  def testFind(self):
    mujoco = parser.from_path(_TEST_MODEL_XML, resolve_references=False)
    mujoco.model = 'model'

    submujoco = copy.copy(mujoco)
    submujoco.model = 'submodel'

    subsubmujoco = copy.copy(mujoco)
    subsubmujoco.model = 'subsubmodel'

    submujoco.find('site', 'attachment').attach(subsubmujoco)
    mujoco.attach(submujoco)

    self.assertIsNotNone(mujoco.find('geom', 'b_0_0'))
    self.assertIsNotNone(mujoco.find('body', 'b_0').find('geom', 'b_0_0'))
    self.assertIsNone(mujoco.find('body', 'b_1').find('geom', 'b_0_0'))

    self.assertIsNone(mujoco.find('geom', 'nonexistent'))
    self.assertIsNone(mujoco.find('geom', 'nonexistent/b_0_0'))

    self.assertEqual(mujoco.find('geom', 'submodel/b_0_0'),
                     submujoco.find('geom', 'b_0_0'))
    self.assertEqual(mujoco.find('geom', 'submodel/subsubmodel/b_0_0'),
                     submujoco.find('geom', 'subsubmodel/b_0_0'))
    self.assertEqual(submujoco.find('geom', 'subsubmodel/b_0_0'),
                     subsubmujoco.find('geom', 'b_0_0'))

    subsubmujoco.find('geom', 'b_0_0').name = 'foo'
    self.assertIsNone(mujoco.find('geom', 'submodel/subsubmodel/b_0_0'))
    self.assertIsNone(submujoco.find('geom', 'subsubmodel/b_0_0'))
    self.assertIsNone(subsubmujoco.find('geom', 'b_0_0'))
    self.assertEqual(mujoco.find('geom', 'submodel/subsubmodel/foo'),
                     submujoco.find('geom', 'subsubmodel/foo'))
    self.assertEqual(submujoco.find('geom', 'subsubmodel/foo'),
                     subsubmujoco.find('geom', 'foo'))

    self.assertEqual(mujoco.find('actuator', 'b_0_0').root, mujoco)
    self.assertEqual(mujoco.find('actuator', 'b_0_0').tag, 'velocity')
    self.assertEqual(mujoco.find('actuator', 'b_0_0').joint, 'b_0_0')

    self.assertEqual(mujoco.find('actuator', 'submodel/b_0_0').root, submujoco)
    self.assertEqual(mujoco.find('actuator', 'submodel/b_0_0').tag, 'velocity')
    self.assertEqual(mujoco.find('actuator', 'submodel/b_0_0').joint, 'b_0_0')

  def testFindInvalidNamespace(self):
    mjcf_model = mjcf.RootElement()
    with six.assertRaisesRegex(self, ValueError, 'not a valid namespace'):
      mjcf_model.find('jiont', 'foo')
    with six.assertRaisesRegex(self, ValueError, 'not a valid namespace'):
      mjcf_model.find_all('goem')

  def testEnterScope(self):
    mujoco = parser.from_path(_TEST_MODEL_XML, resolve_references=False)
    mujoco.model = 'model'

    self.assertIsNone(mujoco.enter_scope('submodel'))

    submujoco = copy.copy(mujoco)
    submujoco.model = 'submodel'

    subsubmujoco = copy.copy(mujoco)
    subsubmujoco.model = 'subsubmodel'

    submujoco.find('site', 'attachment').attach(subsubmujoco)
    mujoco.attach(submujoco)

    self.assertIsNotNone(mujoco.enter_scope('submodel'))
    self.assertEqual(mujoco.enter_scope('submodel').find('geom', 'b_0_0'),
                     submujoco.find('geom', 'b_0_0'))

    self.assertEqual(
        mujoco.enter_scope('submodel/subsubmodel/').find('geom', 'b_0_0'),
        subsubmujoco.find('geom', 'b_0_0'))
    self.assertEqual(mujoco.enter_scope('submodel').enter_scope(
        'subsubmodel').find('geom', 'b_0_0'),
                     subsubmujoco.find('geom', 'b_0_0'))

    self.assertEqual(
        mujoco.enter_scope('submodel').find('actuator', 'b_0_0').root,
        submujoco)
    self.assertEqual(
        mujoco.enter_scope('submodel').find('actuator', 'b_0_0').tag,
        'velocity')
    self.assertEqual(
        mujoco.enter_scope('submodel').find('actuator', 'b_0_0').joint,
        'b_0_0')

  def testDefaultIdentifier(self):
    mujoco = element.RootElement(model='test')
    body = mujoco.worldbody.add('body')
    joint_0 = body.add('freejoint')
    joint_1 = body.add('joint', type='hinge')
    self.assertIsNone(body.name)
    self.assertIsNone(joint_0.name)
    self.assertIsNone(joint_1.name)
    self.assertEqual(str(body), '<body>...</body>')
    self.assertEqual(str(joint_0), '<freejoint/>')
    self.assertEqual(str(joint_1), '<joint class="/" type="hinge"/>')
    self.assertEqual(body.full_identifier, '//unnamed_body_0')
    self.assertStartsWith(body.to_xml_string(pretty_print=False),
                          '<body name="{:s}">'.format(body.full_identifier))
    self.assertEqual(joint_0.full_identifier, '//unnamed_joint_0')
    self.assertEqual(joint_0.to_xml_string(pretty_print=False),
                     '<freejoint name="{:s}"/>'.format(joint_0.full_identifier))
    self.assertEqual(joint_1.full_identifier, '//unnamed_joint_1')
    self.assertEqual(joint_1.to_xml_string(pretty_print=False),
                     '<joint name="{:s}" class="/" type="hinge"/>'.format(
                         joint_1.full_identifier))

    submujoco = copy.copy(mujoco)
    submujoco.model = 'submodel'
    mujoco.attach(submujoco)
    submujoco_body = submujoco.worldbody.body[0]
    self.assertEqual(submujoco_body.full_identifier,
                     'submodel//unnamed_body_0')
    self.assertEqual(submujoco_body.freejoint.full_identifier,
                     'submodel//unnamed_joint_0')
    self.assertEqual(submujoco_body.joint[0].full_identifier,
                     'submodel//unnamed_joint_1')

  def testFindAll(self):
    mujoco = parser.from_path(_TEST_MODEL_XML)
    mujoco.model = 'model'

    submujoco = copy.copy(mujoco)
    submujoco.model = 'submodel'

    subsubmujoco = copy.copy(mujoco)
    subsubmujoco.model = 'subsubmodel'

    submujoco.find('site', 'attachment').attach(subsubmujoco)
    mujoco.attach(submujoco)

    geoms = mujoco.find_all('geom')
    self.assertLen(geoms, 6)
    self.assertEqual(geoms[0].root, mujoco)
    self.assertEqual(geoms[1].root, mujoco)
    self.assertEqual(geoms[2].root, submujoco)
    self.assertEqual(geoms[3].root, subsubmujoco)
    self.assertEqual(geoms[4].root, subsubmujoco)
    self.assertEqual(geoms[5].root, submujoco)

    b_0 = submujoco.find('body', 'b_0')
    self.assertLen(b_0.find_all('joint'), 6)
    self.assertLen(b_0.find_all('joint', immediate_children_only=True), 1)
    self.assertLen(b_0.find_all('joint', exclude_attachments=True), 2)

  def testFindAllFrameJoints(self):
    root_model = parser.from_path(_TEST_MODEL_XML)
    root_model.model = 'model'

    submodel = copy.copy(root_model)
    submodel.model = 'submodel'

    frame = root_model.attach(submodel)
    joint_x = frame.add('joint', type='slide', axis=[1, 0, 0])
    joint_y = frame.add('joint', type='slide', axis=[0, 1, 0])

    joints = frame.find_all('joint', immediate_children_only=True)
    self.assertListEqual(joints, [joint_x, joint_y])

  def testDictLikeInterface(self):
    mujoco = element.RootElement(model='test')
    elem = mujoco.worldbody.add('body')
    if six.PY3:
      subscript_error_regex = 'object is not subscriptable'
    else:
      subscript_error_regex = 'no attribute \'__getitem__\''
    with six.assertRaisesRegex(self, TypeError, subscript_error_regex):
      _ = elem['foo']
    with six.assertRaisesRegex(
        self, TypeError, 'does not support item assignment'):
      elem['foo'] = 'bar'
    with six.assertRaisesRegex(
        self, TypeError, 'does not support item deletion'):
      del elem['foo']

  def testSetAndGetAttributes(self):
    mujoco = element.RootElement(model='test')

    foo_attribs = dict(name='foo', pos=[1, 2, 3, 4], quat=[0, 1, 0, 0])
    with six.assertRaisesRegex(self, ValueError, 'no more than 3 entries'):
      foo = mujoco.worldbody.add('body', **foo_attribs)

    # failed creationg should not cause the identifier 'foo' to be registered
    with self.assertRaises(KeyError):
      mujoco.namescope.get('body', 'foo')
    foo_attribs['pos'] = [1, 2, 3]
    foo = mujoco.worldbody.add('body', **foo_attribs)
    self._test_attributes(foo, expected_values=foo_attribs)

    foo_attribs['name'] = 'bar'
    foo_attribs['pos'] = [1, 2, 3, 4]
    foo_attribs['childclass'] = 'klass'
    with six.assertRaisesRegex(self, ValueError, 'no more than 3 entries'):
      foo.set_attributes(**foo_attribs)

    # failed assignment should not cause the identifier 'bar' to be registered
    with self.assertRaises(KeyError):
      mujoco.namescope.get('body', 'bar')
    foo_attribs['pos'] = [1, 2, 3]
    foo.set_attributes(**foo_attribs)
    self._test_attributes(foo, expected_values=foo_attribs)

    actual_foo_attribs = foo.get_attributes()
    for attribute_name, value in six.iteritems(foo_attribs):
      np.testing.assert_array_equal(
          actual_foo_attribs.pop(attribute_name), value)
    for value in six.itervalues(actual_foo_attribs):
      self.assertIsNone(value)

  def testResolveReferences(self):
    resolved_model = parser.from_path(_TEST_MODEL_XML)
    self.assertIs(
        resolved_model.find('geom', 'b_1_0').material,
        resolved_model.find('material', 'mat_texture'))
    unresolved_model = parser.from_path(
        _TEST_MODEL_XML, resolve_references=False)
    self.assertEqual(
        unresolved_model.find('geom', 'b_1_0').material, 'mat_texture')
    unresolved_model.resolve_references()
    self.assertIs(
        unresolved_model.find('geom', 'b_1_0').material,
        unresolved_model.find('material', 'mat_texture'))

  @parameterized.named_parameters(
      ('WithoutInclude', _TEST_MODEL_XML),
      ('WithInclude', _MODEL_WITH_INCLUDE_PATH))
  def testParseFromString(self, model_path):
    with open(model_path) as xml_file:
      xml_string = xml_file.read()
    model_dir, _ = os.path.split(model_path)
    parser.from_xml_string(xml_string, model_dir=model_dir)

  @parameterized.named_parameters(
      ('WithoutInclude', _TEST_MODEL_XML),
      ('WithInclude', _MODEL_WITH_INCLUDE_PATH))
  def testParseFromFile(self, model_path):
    model_dir, _ = os.path.split(model_path)
    with open(model_path) as xml_file:
      parser.from_file(xml_file, model_dir=model_dir)

  @parameterized.named_parameters(
      ('WithoutInclude', _TEST_MODEL_XML),
      ('WithInclude', _MODEL_WITH_INCLUDE_PATH))
  def testParseFromPath(self, model_path):
    parser.from_path(model_path)

  def testGetAssetFromFile(self):
    with open(_TEXTURE_PATH, 'rb') as f:
      contents = f.read()
    _, filename = os.path.split(_TEXTURE_PATH)
    prefix, extension = os.path.splitext(filename)
    vfs_filename = prefix + '-' + hashlib.sha1(contents).hexdigest() + extension
    mujoco = parser.from_path(_TEST_MODEL_XML)
    self.assertDictEqual({vfs_filename: contents}, mujoco.get_assets())

  def testGetAssetFromPlaceholder(self):
    mujoco = parser.from_path(_TEST_MODEL_XML)
    # Add an extra texture asset from a placeholder.
    contents = b'I am a texture bytestring'
    extension = '.png'
    vfs_filename = hashlib.sha1(contents).hexdigest() + extension
    placeholder = mjcf.Asset(contents=contents, extension=extension)
    mujoco.asset.add('texture', name='fake_texture', file=placeholder)
    self.assertContainsSubset(set([(vfs_filename, contents)]),
                              set(mujoco.get_assets().items()))

  def testGetAssetsFromDict(self):
    with open(_MODEL_WITH_INVALID_FILENAMES, 'rb') as f:
      xml_string = f.read()
    with open(_TEXTURE_PATH, 'rb') as f:
      texture_contents = f.read()
    with open(_MESH_PATH, 'rb') as f:
      mesh_contents = f.read()
    with open(_INCLUDED_WITH_INVALID_FILENAMES, 'rb') as f:
      included_xml_contents = f.read()
    assets = {
        'invalid_texture_name.png': texture_contents,
        'invalid_mesh_name.stl': mesh_contents,
        'invalid_included_name.xml': included_xml_contents,
    }
    # The paths specified in the main and included XML files are deliberately
    # invalid, so the parser should fail unless the pre-loaded assets are passed
    # in as a dict.
    with self.assertRaises(IOError):
      parser.from_xml_string(xml_string=xml_string)

    mujoco = parser.from_xml_string(xml_string=xml_string, assets=assets)
    expected_assets = {}
    for path, contents in six.iteritems(assets):
      _, filename = os.path.split(path)
      prefix, extension = os.path.splitext(filename)
      if extension != '.xml':
        vfs_filename = ''.join(
            [prefix, '-', hashlib.sha1(contents).hexdigest(), extension])
        expected_assets[vfs_filename] = contents
    self.assertDictEqual(expected_assets, mujoco.get_assets())

  def testAssetsCanBeCopied(self):
    with open(_TEXTURE_PATH, 'rb') as f:
      contents = f.read()
    _, filename = os.path.split(_TEXTURE_PATH)
    prefix, extension = os.path.splitext(filename)
    vfs_filename = prefix + '-' + hashlib.sha1(contents).hexdigest() + extension
    mujoco = parser.from_path(_TEST_MODEL_XML)
    mujoco_copy = copy.copy(mujoco)
    expected = {vfs_filename: contents}
    self.assertDictEqual(expected, mujoco.get_assets())
    self.assertDictEqual(expected, mujoco_copy.get_assets())

  def testParseModelWithNamelessAssets(self):
    mujoco = parser.from_path(path=_MODEL_WITH_NAMELESS_ASSETS)
    expected_names_derived_from_filenames = [
        ('mesh', 'cube'),  # ./test_assets/meshes/cube.stl
        ('texture', 'deepmind'),  # ./test_assets/textures/deepmind.png
        ('hfield', 'deepmind'),  # ./test_assets/textures/deepmind.png
    ]
    with self.subTest('Expected asset names are present in the parsed model'):
      for namespace, name in expected_names_derived_from_filenames:
        self.assertIsNotNone(mujoco.find(namespace, name))
    with self.subTest('Can compile and step the simulation'):
      physics = mjcf.Physics.from_mjcf_model(mujoco)
      physics.step()

  def testAssetInheritance(self):
    parent = element.RootElement(model='parent')
    child = element.RootElement(model='child')
    grandchild = element.RootElement(model='grandchild')

    ext = '.png'
    parent_str = b'I belong to the parent'
    child_str = b'I belong to the child'
    grandchild_str = b'I belong to the grandchild'
    parent_vfs_name, child_vfs_name, grandchild_vfs_name = (
        hashlib.sha1(s).hexdigest() + ext
        for s in (parent_str, child_str, grandchild_str))

    parent_ph = mjcf.Asset(contents=parent_str, extension=ext)
    child_ph = mjcf.Asset(contents=child_str, extension=ext)
    grandchild_ph = mjcf.Asset(contents=grandchild_str, extension=ext)

    parent.asset.add('texture', name='parent_tex', file=parent_ph)
    child.asset.add('texture', name='child_tex', file=child_ph)
    grandchild.asset.add('texture', name='grandchild_tex', file=grandchild_ph)

    parent.attach(child)
    child.attach(grandchild)

    # The grandchild should only return its own assets.
    self.assertDictEqual(
        {grandchild_vfs_name: grandchild_str},
        grandchild.get_assets())

    # The child should return its own assets plus those of the grandchild.
    self.assertDictEqual(
        {child_vfs_name: child_str,
         grandchild_vfs_name: grandchild_str},
        child.get_assets())

    # The parent should return everything.
    self.assertDictEqual(
        {parent_vfs_name: parent_str,
         child_vfs_name: child_str,
         grandchild_vfs_name: grandchild_str},
        parent.get_assets())

  def testActuatorReordering(self):

    def make_model_with_mixed_actuators(name):
      actuators = []
      root = mjcf.RootElement(model=name)
      body = root.worldbody.add('body')
      body.add('geom', type='sphere', size=[0.1])
      slider = body.add('joint', type='slide', name='slide_joint')
      # Third-order `general` actuator.
      actuators.append(
          root.actuator.add(
              'general', dyntype='integrator', biastype='affine',
              dynprm=[1, 0, 0], joint=slider, name='general_act'))
      # Cylinder actuators are also third-order.
      actuators.append(
          root.actuator.add('cylinder', joint=slider, name='cylinder_act'))
      # A second-order actuator, added after the third-order actuators.
      actuators.append(
          root.actuator.add('velocity', joint=slider, name='velocity_act'))
      return root, actuators

    child_1, actuators_1 = make_model_with_mixed_actuators(name='child_1')
    child_2, actuators_2 = make_model_with_mixed_actuators(name='child_2')
    child_3, actuators_3 = make_model_with_mixed_actuators(name='child_3')
    parent = mjcf.RootElement()
    parent.attach(child_1)
    parent.attach(child_2)
    child_2.attach(child_3)

    # Check that the generated XML contains all of the actuators that we expect
    # it to have.
    expected_xml_strings = [
        actuator.to_xml_string(prefix_root=parent.namescope)
        for actuator in actuators_1 + actuators_2 + actuators_3
    ]
    xml_strings = [
        util.to_native_string(etree.tostring(node, pretty_print=True))
        for node in parent.to_xml().find('actuator').getchildren()
    ]
    self.assertSameElements(expected_xml_strings, xml_strings)

    # MuJoCo requires that all 3rd-order actuators (i.e. those with internal
    # dynamics) come after all 2nd-order actuators in the XML. Attempting to
    # compile this model will result in an error unless PyMJCF internally
    # reorders the actuators so that the 3rd-order actuator comes last in the
    # generated XML.
    _ = mjcf.Physics.from_mjcf_model(child_1)

    # Actuator re-ordering should also work in cases where there are multiple
    # attached submodels with mixed 2nd- and 3rd-order actuators.
    _ = mjcf.Physics.from_mjcf_model(parent)

  def testMaxConflictingValues(self):
    model_1 = mjcf.RootElement()
    model_1.size.nconmax = 123
    model_1.size.njmax = 456

    model_2 = mjcf.RootElement()
    model_2.size.nconmax = 345
    model_2.size.njmax = 234

    model_1.attach(model_2)
    self.assertEqual(model_1.size.nconmax, 345)
    self.assertEqual(model_1.size.njmax, 456)


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Viewer application module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import _render
from dm_control.viewer import gui
from dm_control.viewer import renderer
from dm_control.viewer import runtime
from dm_control.viewer import user_input
from dm_control.viewer import util
from dm_control.viewer import viewer
from dm_control.viewer import views

_DOUBLE_BUFFERING = (user_input.KEY_F5)
_PAUSE = user_input.KEY_SPACE
_RESTART = user_input.KEY_BACKSPACE
_ADVANCE_SIMULATION = user_input.KEY_RIGHT
_SPEED_UP_TIME = user_input.KEY_EQUAL
_SLOW_DOWN_TIME = user_input.KEY_MINUS
_HELP = user_input.KEY_F1
_STATUS = user_input.KEY_F2

_MAX_FRONTBUFFER_SIZE = 2048
_MISSING_STATUS_ENTRY = '--'
_RUNTIME_STOPPED_LABEL = 'EPISODE TERMINATED - hit backspace to restart'
_STATUS_LABEL = 'Status'
_TIME_LABEL = 'Time'
_CPU_LABEL = 'CPU'
_FPS_LABEL = 'FPS'
_CAMERA_LABEL = 'Camera'
_PAUSED_LABEL = 'Paused'
_ERROR_LABEL = 'Error'


class Help(views.ColumnTextModel):
  """Contains the description of input map employed in the application."""

  def __init__(self):
    """Instance initializer."""
    self._value = [
        ['Help', 'F1'],
        ['Info', 'F2'],
        ['Stereo', 'F5'],
        ['Frame', 'F6'],
        ['Label', 'F7'],
        ['--------------', ''],
        ['Pause', 'Space'],
        ['Reset', 'BackSpace'],
        ['Autoscale', 'Ctrl A'],
        ['Geoms', '0 - 4'],
        ['Sites', 'Shift 0 - 4'],
        ['Speed Up', '='],
        ['Slow Down', '-'],
        ['Switch Cam', '[ ]'],
        ['--------------', ''],
        ['Translate', 'R drag'],
        ['Rotate', 'L drag'],
        ['Zoom', 'Scroll'],
        ['Select', 'L dblclick'],
        ['Center', 'R dblclick'],
        ['Track', 'Ctrl R dblclick / Esc'],
        ['Perturb', 'Ctrl [Shift] L/R drag'],
    ]

  def get_columns(self):
    """Returns the text to display in two columns."""
    return self._value


class Status(views.ColumnTextModel):
  """Monitors and returns the status of the application."""

  def __init__(self, time_multiplier, pause, frame_timer):
    """Instance initializer.

    Args:
      time_multiplier: Instance of util.TimeMultiplier.
      pause: An observable pause subject, instance of util.ObservableFlag.
      frame_timer: A Timer instance counting duration of frames.
    """
    self._runtime = None
    self._time_multiplier = time_multiplier
    self._camera = None
    self._pause = pause
    self._frame_timer = frame_timer
    self._fps_counter = util.Integrator()
    self._cpu_counter = util.Integrator()

    self._value = collections.OrderedDict([
        (_STATUS_LABEL, _MISSING_STATUS_ENTRY),
        (_TIME_LABEL, _MISSING_STATUS_ENTRY),
        (_CPU_LABEL, _MISSING_STATUS_ENTRY),
        (_FPS_LABEL, _MISSING_STATUS_ENTRY),
        (_CAMERA_LABEL, _MISSING_STATUS_ENTRY),
        (_PAUSED_LABEL, _MISSING_STATUS_ENTRY),
        (_ERROR_LABEL, _MISSING_STATUS_ENTRY),
    ])

  def set_camera(self, camera):
    """Updates the active camera instance.

    Args:
      camera: Instance of renderer.SceneCamera.
    """
    self._camera = camera

  def set_runtime(self, instance):
    """Updates the active runtime instance.

    Args:
      instance: Instance of runtime.Base.
    """
    if self._runtime:
      self._runtime.on_error -= self._on_error
      self._runtime.on_episode_begin -= self._clear_error
    self._runtime = instance
    if self._runtime:
      self._runtime.on_error += self._on_error
      self._runtime.on_episode_begin += self._clear_error

  def get_columns(self):
    """Returns the text to display in two columns."""
    if self._frame_timer.measured_time > 0:
      self._fps_counter.value = 1. / self._frame_timer.measured_time
    self._value[_FPS_LABEL] = '{0:.1f}'.format(self._fps_counter.value)

    if self._runtime:
      if self._runtime.state == runtime.State.STOPPED:
        self._value[_STATUS_LABEL] = _RUNTIME_STOPPED_LABEL
      else:
        self._value[_STATUS_LABEL] = str(self._runtime.state)

      self._cpu_counter.value = self._runtime.simulation_time

      self._value[_TIME_LABEL] = '{0:.1f} ({1}x)'.format(
          self._runtime.get_time(), str(self._time_multiplier))
      self._value[_CPU_LABEL] = '{0:.2f}ms'.format(
          self._cpu_counter.value * 1000.0)
    else:
      self._value[_STATUS_LABEL] = _MISSING_STATUS_ENTRY
      self._value[_TIME_LABEL] = _MISSING_STATUS_ENTRY
      self._value[_CPU_LABEL] = _MISSING_STATUS_ENTRY

    if self._camera:
      self._value[_CAMERA_LABEL] = self._camera.name
    else:
      self._value[_CAMERA_LABEL] = _MISSING_STATUS_ENTRY

    self._value[_PAUSED_LABEL] = str(self._pause.value)

    return list(self._value.items())  # For Python 2/3 compatibility.

  def _clear_error(self):
    self._value[_ERROR_LABEL] = _MISSING_STATUS_ENTRY

  def _on_error(self, error_msg):
    self._value[_ERROR_LABEL] = error_msg


class ReloadParams(collections.namedtuple(
    'RefreshParams', ['zoom_to_scene'])):
  """Parameters of a reload request."""


class Application(object):
  """Viewer application."""

  def __init__(self, title='Explorer', width=1024, height=768):
    """Instance initializer."""
    self._render_surface = None
    self._renderer = renderer.NullRenderer()
    self._viewport = renderer.Viewport(width, height)
    self._window = gui.RenderWindow(width, height, title)

    self._pause_subject = util.ObservableFlag(True)
    self._time_multiplier = util.TimeMultiplier(1.)
    self._frame_timer = util.Timer()
    self._viewer = viewer.Viewer(
        self._viewport, self._window.mouse, self._window.keyboard)
    self._viewer_layout = views.ViewportLayout()
    self._status = Status(
        self._time_multiplier, self._pause_subject, self._frame_timer)

    self._runtime = None
    self._environment_loader = None
    self._environment = None
    self._policy = None
    self._deferred_reload_request = None

    status_view_toggle = self._build_view_toggle(
        views.ColumnTextView(self._status), views.PanelLocation.BOTTOM_LEFT)
    help_view_toggle = self._build_view_toggle(
        views.ColumnTextView(Help()), views.PanelLocation.TOP_RIGHT)
    status_view_toggle()

    self._input_map = user_input.InputMap(
        self._window.mouse, self._window.keyboard)
    self._input_map.bind(self._pause_subject.toggle, _PAUSE)
    self._input_map.bind(self._time_multiplier.increase, _SPEED_UP_TIME)
    self._input_map.bind(self._time_multiplier.decrease, _SLOW_DOWN_TIME)
    self._input_map.bind(self._advance_simulation, _ADVANCE_SIMULATION)
    self._input_map.bind(self._restart_runtime, _RESTART)
    self._input_map.bind(help_view_toggle, _HELP)
    self._input_map.bind(status_view_toggle, _STATUS)

  def _on_reload(self, zoom_to_scene=False):
    """Perform initialization related to Physics reload.

    Reset the components that depend on a specific Physics class instance.

    Args:
      zoom_to_scene: Should the camera zoom to show the entire scene after the
        reload is complete.
    """
    self._deferred_reload_request = ReloadParams(zoom_to_scene)
    self._viewer.deinitialize()
    self._status.set_camera(None)

  def _perform_deferred_reload(self, params):
    """Performs the deferred part of initialization related to Physics reload.

    Args:
      params: Deferred reload parameters, an instance of ReloadParams.
    """
    if self._render_surface:
      self._render_surface.free()
    if self._renderer:
      self._renderer.release()
    self._render_surface = _render.Renderer(
        max_width=_MAX_FRONTBUFFER_SIZE, max_height=_MAX_FRONTBUFFER_SIZE)
    self._renderer = renderer.OffScreenRenderer(
        self._environment.physics.model, self._render_surface)
    self._renderer.components += self._viewer_layout
    self._viewer.initialize(
        self._environment.physics, self._renderer, touchpad=False)
    self._status.set_camera(self._viewer.camera)
    if params.zoom_to_scene:
      self._viewer.zoom_to_scene()

  def _build_view_toggle(self, view, location):
    def toggle():
      if view in self._viewer_layout:
        self._viewer_layout.remove(view)
      else:
        self._viewer_layout.add(view, location)
    return toggle

  def _tick(self):
    """Handle GUI events until the main window is closed."""
    if self._deferred_reload_request:
      self._perform_deferred_reload(self._deferred_reload_request)
      self._deferred_reload_request = None
    time_elapsed = self._frame_timer.tick() * self._time_multiplier.get()
    if self._runtime:
      with self._viewer.perturbation.apply(self._pause_subject.value):
        self._runtime.tick(time_elapsed, self._pause_subject.value)
    self._viewer.render()

  def _load_environment(self, zoom_to_scene):
    """Loads a new environment."""
    if self._runtime:
      del self._runtime
      self._runtime = None
    self._environment = None
    environment_instance = None
    if self._environment_loader:
      environment_instance = self._environment_loader()
    if environment_instance:
      self._environment = environment_instance
      self._runtime = runtime.Runtime(
          environment=self._environment, policy=self._policy)
      self._runtime.on_physics_changed += lambda: self._on_reload(False)
    self._status.set_runtime(self._runtime)
    self._on_reload(zoom_to_scene=zoom_to_scene)

  def _restart_runtime(self):
    """Restarts the episode, resetting environment, model, and data."""
    if self._runtime:
      self._runtime.stop()
    self._load_environment(zoom_to_scene=False)

  def _advance_simulation(self):
    if self._runtime:
      self._runtime.single_step()

  def launch(self, environment_loader, policy=None):
    """Starts the viewer with the specified policy and environment.

    Args:
      environment_loader: Either a callable that takes no arguments and returns
        an instance of dm_control.rl.control.Environment, or an instance of
        dm_control.rl.control.Environment.
      policy: An optional callable corresponding to a policy to execute
        within the environment. It should accept a `TimeStep` and return
        a numpy array of actions conforming to the output of
        `environment.action_spec()`.

    Raises:
      ValueError: If `environment_loader` is None.
    """
    if environment_loader is None:
      raise ValueError('"environment_loader" argument is required.')
    if callable(environment_loader):
      self._environment_loader = environment_loader
    else:
      self._environment_loader = lambda: environment_loader
    self._policy = policy
    self._load_environment(zoom_to_scene=True)
    def tick():
      self._viewport.set_size(*self._window.shape)
      self._tick()
      return self._renderer.pixels
    self._window.event_loop(tick_func=tick)
    self._window.close()



# Copyright 2017-2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for inverse_kinematics."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools

from absl.testing import absltest
from absl.testing import parameterized
from dm_control import mujoco
from dm_control.mujoco.testing import assets
from dm_control.mujoco.wrapper import mjbindings
from dm_control.utils import inverse_kinematics as ik
import numpy as np

mjlib = mjbindings.mjlib

_ARM_XML = assets.get_contents('arm.xml')
_MODEL_WITH_BALL_JOINTS_XML = assets.get_contents('model_with_ball_joints.xml')

_SITE_NAME = 'gripsite'
_JOINTS = ['joint_1', 'joint_2', 'joint_3', 'joint_4', 'joint_5', 'joint_6']
_TOL = 1e-14
_MAX_STEPS = 100
_MAX_RESETS = 10

_TARGETS = [
    # target_pos              # target_quat
    (np.array([0., 0., 0.3]), np.array([0., 1., 0., 1.])),
    (np.array([-0.5, 0., 0.5]), None),
    (np.array([0., 0., 0.8]), np.array([0., 1., 0., 1.])),
    (np.array([0., 0., 0.8]), None),
    (np.array([0., -0.1, 0.5]), None),
    (np.array([0., -0.1, 0.5]), np.array([1., 1., 0., 0.])),
    (np.array([0.5, 0., 0.5]), None),
    (np.array([0.4, 0.1, 0.5]), None),
    (np.array([0.4, 0.1, 0.5]), np.array([1., 0., 0., 0.])),
    (np.array([0., 0., 0.3]), None),
    (np.array([0., 0.5, -0.2]), None),
    (np.array([0.5, 0., 0.3]), np.array([1., 0., 0., 1.])),
    (None, np.array([1., 0., 0., 1.])),
    (None, np.array([0., 1., 0., 1.])),
]
_INPLACE = [False, True]


class _ResetArm(object):

  def __init__(self, seed=None):
    self._rng = np.random.RandomState(seed)
    self._lower = None
    self._upper = None

  def _cache_bounds(self, physics):
    self._lower, self._upper = physics.named.model.jnt_range[_JOINTS].T
    limited = physics.named.model.jnt_limited[_JOINTS].astype(np.bool)
    # Positions for hinge joints without limits are sampled between 0 and 2pi
    self._lower[~limited] = 0
    self._upper[~limited] = 2 * np.pi

  def __call__(self, physics):
    if self._lower is None:
      self._cache_bounds(physics)
    # NB: This won't work for joints with > 1 DOF
    new_qpos = self._rng.uniform(self._lower, self._upper)
    physics.named.data.qpos[_JOINTS] = new_qpos


class InverseKinematicsTest(parameterized.TestCase):

  @parameterized.parameters(itertools.product(_TARGETS, _INPLACE))
  def testQposFromSitePose(self, target, inplace):
    physics = mujoco.Physics.from_xml_string(_ARM_XML)
    target_pos, target_quat = target
    count = 0
    physics2 = physics.copy(share_model=True)
    resetter = _ResetArm(seed=0)
    while True:
      result = ik.qpos_from_site_pose(
          physics=physics2,
          site_name=_SITE_NAME,
          target_pos=target_pos,
          target_quat=target_quat,
          joint_names=_JOINTS,
          tol=_TOL,
          max_steps=_MAX_STEPS,
          inplace=inplace,
      )
      if result.success:
        break
      elif count < _MAX_RESETS:
        resetter(physics2)
        count += 1
      else:
        raise RuntimeError(
            'Failed to find a solution within %i attempts.' % _MAX_RESETS)

    self.assertLessEqual(result.steps, _MAX_STEPS)
    self.assertLessEqual(result.err_norm, _TOL)
    physics.data.qpos[:] = result.qpos
    mjlib.mj_fwdPosition(physics.model.ptr, physics.data.ptr)
    if target_pos is not None:
      pos = physics.named.data.site_xpos[_SITE_NAME]
      np.testing.assert_array_almost_equal(pos, target_pos)
    if target_quat is not None:
      xmat = physics.named.data.site_xmat[_SITE_NAME]
      quat = np.empty_like(target_quat)
      mjlib.mju_mat2Quat(quat, xmat)
      quat /= quat.ptp()  # Normalize xquat so that its max-min range is 1
      np.testing.assert_array_almost_equal(quat, target_quat)

  def testNamedJointsWithMultipleDOFs(self):
    """Regression test for b/77506142."""
    physics = mujoco.Physics.from_xml_string(_MODEL_WITH_BALL_JOINTS_XML)
    site_name = 'gripsite'
    joint_names = ['joint_1', 'joint_2']

    # This target position can only be achieved by rotating both ball joints. If
    # DOFs are incorrectly indexed by joint index then only the first two DOFs
    # in the first ball joint can change, and IK will fail to find a solution.
    target_pos = (0.05, 0.05, 0)
    result = ik.qpos_from_site_pose(
        physics=physics,
        site_name=site_name,
        target_pos=target_pos,
        joint_names=joint_names,
        tol=_TOL,
        max_steps=_MAX_STEPS,
        inplace=True)

    self.assertTrue(result.success)
    self.assertLessEqual(result.steps, _MAX_STEPS)
    self.assertLessEqual(result.err_norm, _TOL)
    pos = physics.named.data.site_xpos[site_name]
    np.testing.assert_array_almost_equal(pos, target_pos)

    # IK should fail to converge if only the first joint is allowed to move.
    physics.reset()
    result = ik.qpos_from_site_pose(
        physics=physics,
        site_name=site_name,
        target_pos=target_pos,
        joint_names=joint_names[:1],
        tol=_TOL,
        max_steps=_MAX_STEPS,
        inplace=True)
    self.assertFalse(result.success)

  @parameterized.named_parameters(
      ('None', None),
      ('list', ['joint_1', 'joint_2']),
      ('tuple', ('joint_1', 'joint_2')),
      ('numpy.array', np.array(['joint_1', 'joint_2'])))
  def testAllowedJointNameTypes(self, joint_names):
    """Test allowed types for joint_names parameter."""
    physics = mujoco.Physics.from_xml_string(_ARM_XML)
    site_name = 'gripsite'
    target_pos = (0.05, 0.05, 0)
    ik.qpos_from_site_pose(
        physics=physics,
        site_name=site_name,
        target_pos=target_pos,
        joint_names=joint_names,
        tol=_TOL,
        max_steps=_MAX_STEPS,
        inplace=True)

  @parameterized.named_parameters(
      ('int', 1),
      ('dict', {'joint_1': 1, 'joint_2': 2}),
      ('function', lambda x: x),
  )
  def testDisallowedJointNameTypes(self, joint_names):
    physics = mujoco.Physics.from_xml_string(_ARM_XML)
    site_name = 'gripsite'
    target_pos = (0.05, 0.05, 0)

    expected_message = ik._INVALID_JOINT_NAMES_TYPE.format(type(joint_names))

    with self.assertRaisesWithLiteralMatch(ValueError, expected_message):
      ik.qpos_from_site_pose(
          physics=physics,
          site_name=site_name,
          target_pos=target_pos,
          joint_names=joint_names,
          tol=_TOL,
          max_steps=_MAX_STEPS,
          inplace=True)

  def testNoTargetPosOrQuat(self):
    physics = mujoco.Physics.from_xml_string(_ARM_XML)
    site_name = 'gripsite'
    with self.assertRaisesWithLiteralMatch(
        ValueError, ik._REQUIRE_TARGET_POS_OR_QUAT):
      ik.qpos_from_site_pose(
          physics=physics,
          site_name=site_name,
          tol=_TOL,
          max_steps=_MAX_STEPS,
          inplace=True)

if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Corridor-based locomotion tasks."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control import composer
from dm_control.composer import variation
from dm_control.utils import rewards
import numpy as np


class RunThroughCorridor(composer.Task):
  """A task that requires a walker to run through a corridor.

  This task rewards an agent for controlling a walker to move at a specific
  target velocity along the corridor, and for minimising the magnitude of the
  control signals used to achieve this.
  """

  def __init__(self,
               walker,
               arena,
               walker_spawn_position=(0, 0, 0),
               walker_spawn_rotation=None,
               target_velocity=3.0,
               contact_termination=True,
               terminate_at_height=-0.5,
               physics_timestep=0.005,
               control_timestep=0.025):
    """Initializes this task.

    Args:
      walker: an instance of `locomotion.walkers.base.Walker`.
      arena: an instance of `locomotion.arenas.corridors.Corridor`.
      walker_spawn_position: a sequence of 3 numbers, or a `composer.Variation`
        instance that generates such sequences, specifying the position at
        which the walker is spawned at the beginning of an episode.
      walker_spawn_rotation: a number, or a `composer.Variation` instance that
        generates a number, specifying the yaw angle offset (in radians) that is
        applied to the walker at the beginning of an episode.
      target_velocity: a number specifying the target velocity (in meters per
        second) for the walker.
      contact_termination: whether to terminate if a non-foot geom touches the
        ground.
      terminate_at_height: a number specifying the height of end effectors below
        which the episode terminates.
      physics_timestep: a number specifying the timestep (in seconds) of the
        physics simulation.
      control_timestep: a number specifying the timestep (in seconds) at which
        the agent applies its control inputs (in seconds).
    """

    self._arena = arena
    self._walker = walker
    self._walker.create_root_joints(self._arena.attach(self._walker))
    self._walker_spawn_position = walker_spawn_position
    self._walker_spawn_rotation = walker_spawn_rotation

    enabled_observables = []
    enabled_observables += self._walker.observables.proprioception
    enabled_observables += self._walker.observables.kinematic_sensors
    enabled_observables += self._walker.observables.dynamic_sensors
    enabled_observables.append(self._walker.observables.sensors_touch)
    enabled_observables.append(self._walker.observables.egocentric_camera)
    for observable in enabled_observables:
      observable.enabled = True

    self._vel = target_velocity
    self._contact_termination = contact_termination
    self._terminate_at_height = terminate_at_height

    self.set_timesteps(
        physics_timestep=physics_timestep, control_timestep=control_timestep)

  @property
  def root_entity(self):
    return self._arena

  def initialize_episode_mjcf(self, random_state):
    self._arena.regenerate(random_state)
    self._arena.mjcf_model.visual.map.znear = 0.00025
    self._arena.mjcf_model.visual.map.zfar = 4.

  def initialize_episode(self, physics, random_state):
    self._walker.reinitialize_pose(physics, random_state)
    if self._walker_spawn_rotation:
      rotation = variation.evaluate(
          self._walker_spawn_rotation, random_state=random_state)
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
    else:
      quat = None
    self._walker.shift_pose(
        physics,
        position=variation.evaluate(
            self._walker_spawn_position, random_state=random_state),
        quaternion=quat,
        rotate_velocity=True)

    self._failure_termination = False
    walker_foot_geoms = set(self._walker.ground_contact_geoms)
    walker_nonfoot_geoms = [
        geom for geom in self._walker.mjcf_model.find_all('geom')
        if geom not in walker_foot_geoms]
    self._walker_nonfoot_geomids = set(
        physics.bind(walker_nonfoot_geoms).element_id)
    self._ground_geomids = set(
        physics.bind(self._arena.ground_geoms).element_id)

  def _is_disallowed_contact(self, contact):
    set1, set2 = self._walker_nonfoot_geomids, self._ground_geomids
    return ((contact.geom1 in set1 and contact.geom2 in set2) or
            (contact.geom1 in set2 and contact.geom2 in set1))

  def before_step(self, physics, action, random_state):
    self._walker.apply_action(physics, action, random_state)

  def after_step(self, physics, random_state):
    self._failure_termination = False
    if self._contact_termination:
      for c in physics.data.contact:
        if self._is_disallowed_contact(c):
          self._failure_termination = True
          break
    if self._terminate_at_height is not None:
      if any(physics.bind(self._walker.end_effectors).xpos[:, -1] <
             self._terminate_at_height):
        self._failure_termination = True

  def get_reward(self, physics):
    walker_xvel = physics.bind(self._walker.root_body).subtree_linvel[0]
    xvel_term = rewards.tolerance(
        walker_xvel, (self._vel, self._vel),
        margin=self._vel,
        sigmoid='linear',
        value_at_margin=0.0)
    return xvel_term

  def should_terminate_episode(self, physics):
    return self._failure_termination

  def get_discount(self, physics):
    if self._failure_termination:
      return 0.
    else:
      return 1.



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Cameras for recording soccer videos."""

from dm_control.mujoco import engine
import numpy as np


class MultiplayerTrackingCamera(object):
  """Camera that smoothly tracks multiple entities."""

  def __init__(
      self,
      min_distance,
      distance_factor,
      smoothing_update_speed,
      azimuth=90,
      elevation=-45,
      width=1920,
      height=1080,
  ):
    """Construct a new MultiplayerTrackingcamera.

    The target lookat point is the centroid of all tracked entities.
    Target camera distance is set to min_distance + distance_factor * d_max,
    where d_max is the maximum distance of any entity to the lookat point.

    Args:
      min_distance: minimum camera distance.
      distance_factor: camera distance multiplier (see above).
      smoothing_update_speed: exponential filter parameter to smooth camera
        movement. 1 means no filter; smaller values mean less change per step.
      azimuth: constant azimuth to use for camera.
      elevation: constant elevation to use for camera.
      width: width to use for rendered video.
      height: height to use for rendered video.
    """
    self._min_distance = min_distance
    self._distance_factor = distance_factor
    if smoothing_update_speed < 0 or smoothing_update_speed > 1:
      raise ValueError("Filter speed must be in range [0, 1].")
    self._smoothing_update_speed = smoothing_update_speed
    self._azimuth = azimuth
    self._elevation = elevation
    self._width = width
    self._height = height
    self._camera = None

  @property
  def camera(self):
    return self._camera

  def render(self):
    """Render the current frame."""
    if self._camera is None:
      raise ValueError(
          "Camera has not been initialized yet."
          " render can only be called after physics has been compiled."
      )
    return self._camera.render()

  def after_compile(self, physics):
    """Instantiate the camera and ensure rendering buffer is large enough."""
    buffer_height = max(self._height, physics.model.vis.global_.offheight)
    buffer_width = max(self._width, physics.model.vis.global_.offwidth)
    physics.model.vis.global_.offheight = buffer_height
    physics.model.vis.global_.offwidth = buffer_width
    self._camera = engine.MovableCamera(
        physics, height=self._height, width=self._width)

  def _get_target_camera_pose(self, entity_positions):
    """Returns the pose that the camera should be pulled toward.

    Args:
      entity_positions: list of numpy arrays representing current positions of
        the entities to be tracked.
    Returns: mujoco.engine.Pose representing the target camera pose.
    """
    stacked_positions = np.stack(entity_positions)
    centroid = np.mean(stacked_positions, axis=0)
    radii = np.linalg.norm(stacked_positions - centroid, axis=1)
    assert len(radii) == len(entity_positions)
    camera_distance = self._min_distance + self._distance_factor * np.max(radii)
    return engine.Pose(
        lookat=centroid,
        distance=camera_distance,
        azimuth=self._azimuth,
        elevation=self._elevation,
    )

  def initialize_episode(self, entity_positions):
    """Begin the episode with the camera set to its target pose."""
    target_pose = self._get_target_camera_pose(entity_positions)
    self._camera.set_pose(*target_pose)

  def after_step(self, entity_positions):
    """Move camera toward its target poses."""
    target_pose = self._get_target_camera_pose(entity_positions)
    cur_pose = self._camera.get_pose()
    smoothing_update_speed = self._smoothing_update_speed
    filtered_pose = [
        target_val * smoothing_update_speed + \
        current_val * (1 - smoothing_update_speed)
        for target_val, current_val in zip(target_pose, cur_pose)
    ]
    self._camera.set_pose(*filtered_pose)
